#!/usr/bin/env python
"""
PYTHON SCRIPT TO INTERACT WITH PERMA LEXICON DBs

TODO:
-make sure printCSV covers all bases.
-Add fucntionaltiy for adding terms to db
-make sure add terms works with hashtags
"""
import math
import sys,os, getpass
import re, csv
import pprint
import time
import MySQLdb
import random

#from optparse import OptionParser, OptionGroup
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
try:
    from nltk.corpus import wordnet as wn
except ImportError:
    print('LexInterface:warning: nltk.corpus module not imported.')

sys.path.append(os.path.dirname(os.path.realpath(__file__)).replace("/dlatk/LexicaInterface",""))
import dlatk.dlaConstants as dlac
from dlatk.mysqlmethods.mysqlMethods import abstractDBConnect, dbConnect

##CONSTANTS (STATIC VARIABLES)##
PERMA_CODES = {'P+': 'positive emotion',
              'P-': 'negative emotion',
              'E+': 'engagement',
              'E-': 'disengagement',
              'R+': 'positive relationships',
              'R-': 'negative relationships',
              'M+': 'meaning in life',
              'M-': 'meaninglessness',
              'A+': 'accomplishment',
              'A-': 'unaccomplishment'
              }

PERMA_SHORT_DEF = {'P+': 'positive feelings such as happiness, joy, excitement, enthusiasm, and contentment',
                   'P-': 'negative feelings such as distress, anger, contempt, disgust, fear, nervousness, and sadness',
                   'E+': 'being absorbed, focused, or interested in what one is doing',
                   'E-': 'withdrawal from life, lack of activity, disinterest, or boredom',
                   'R+': 'caring for others and feeling cared for, loved, esteemed, and valued by others',
                   'R-': 'disconnected from others, loneliness, or feeling like their relationships with other people are bad for them',
                   'M+': 'feeling of belonging or serving something larger than the self, having a sense of purpose or direction',
                   'M-': 'feeling disconnected, worthless, drifting, like their life is without a purpose',
                   'A+': 'achievement, success, or feeling like they mastered something',
                   'A-': 'lack of achievement or mastery, settling for status quo or mediocracy'
              }
PERMA_LONG_DEF = {'P+': 'If we placed all emotions on a spectrum ranging from pleasant to unpleasant, positive emotions are those emotions on the pleasant end that people prefer to feel (e.g., feeling grateful, upbeat; expressing appreciation, liking). Positive emotions can be intense, or fleeting (i.e. excitement, ecstasy, or joy) or they can be milder, and longer lasting (i.e. gratitude, contentment, or peaceful).',
              'P-': 'Negative emotions represent unpleasant feelings, that people would prefer not to feel (e.g., feeling contemptuous, irritable; expressing disdain, disliking). Negative emotions can be primal and intense (i.e. rage, jealousy, or depression) or they can be milder (i.e. annoyance, dislike, or sadness).',
              'E+': 'When one is engaged in their life, work, or activity, the individual is completely absorbed, interested and focused on what they are doing. The extreme form of engagement is sometimes called "flow." In flow, time passes quickly, attention is completely focused on the activity and the sense of self is lost. The message can be in the past tense .',
              'E-': 'When one is not feeling engaged in their life, work, or activity the individual can be feeling bored, uninterested, sleepy or unable to concentrate on the present moment. When rating messages, for a high rating in unengagement the message would undoubtedly be expressing strong unengagement in their activity and most of the message would be about expressing that unengagement. The message can be in the past tense.',
              'R+': 'Whe one feels that relationships are good, making them feel loved, respected and valued by the others in their lives or having positive feelings towards others in their lives.  One might express feeling grateful and loving for the people in the in their lives.',
              'R-': 'When one feels lonely, disconnected from others, judgmental of others or generally feeling like they have few or no good relationships in their lives or that the relationships. One may also express dissatisfaction with their current relationships.',
              'M+': 'When one generally feels like they have a sense of direction in their lives. One may express being "called" to do something or like they have set life plan. One might talk about their religion or spirituality or they may talk about their fate and destiny. One may also talk about particular events happening for a reason.',
              'M-': 'When one generally feels direction-less in life or like their life or life events lack any higher meaning. One may express that their lives are pointless or may talk about particular life events are pointless and server no higher reason.',
              'A+': 'When one feels like they are accomplishing something or like they are moving forward and advancing in their life. One may express feeling proud of doing something or proud of someone else\'s accomplishments.',
              'A-': 'When one feels like they are stuck, or not accomplishing anything with their life. One may express feeling disappointed with themselves or like they\'re not measuring up.'
              }

############################################################
## Class / Static Methods

def loadLexiconFromFile(filename):
    """Loads the perma lexicon, using standard formatting
    returns a dictionary of frozensets"""
    lexFile = open(filename)
    reader = csv.reader(lexFile,dialect='excel')
    # comma = re.compile(r'\,\s*')
    cats = []
    lexicon = {}
    # for line in lexFile:
    for terms in reader:
        # terms = comma.split(line.rstrip())
        if len(cats) > 0:
            for i in range(len(terms)):
                if (terms[i]):
                    lexicon[cats[i]].append(terms[i])
        else:
            for i in range(len(terms)):
                lexicon[terms[i]] = []
                cats.append(terms[i])
    
    for cat in cats:
        lexicon[cat] = frozenset(lexicon[cat])

    return lexicon

#LUKASZ EDIT
def loadLexiconFromGFile(filename, using_filter):
    """Loads a lexicon in "google format"
    returns a dictionary of frozensets"""
    lexFile = open(filename)
    lexLines = lexFile.read().split('\r')
    lexLines = lexLines[1:len(lexLines)]

    comma = re.compile(r'\,')
    cats = []
    lexicon = {}
    for line in lexLines:
        line_split = comma.split(line)

        if len(line_split)>1:
            this_cat = line_split[1]
            this_term = line_split[0]
            this_keep = int(line_split[2])
            if this_cat in lexicon:
                if using_filter:
                    if this_keep == 1:
                        lexicon[this_cat].append(this_term)
                else:
                    lexicon[this_cat].append(this_term)
            else:
                lexicon[this_cat] = []
                cats.append(this_cat)

    for cat in cats:
        lexicon[cat] = frozenset(lexicon[cat])

    return lexicon    

def loadLexiconFromSparse(filename):
    """Loads the perma lexicon from a sparse formatting word[, word], category
    returns a dictionary of frozensets"""
    lexFile = open(filename)
    comma = re.compile(r'\,\s*')
    cats = []
    lexicon = {}
    for line in lexFile:
        items = comma.split(line.rstrip())
        terms = items[0:-1]
        cat = items[-1]
        if cat not in lexicon:
            lexicon[cat] = []
        for term in terms:
            if term: 
                lexicon[cat].append(term)
    
    for cat in list(lexicon.keys()):
        lexicon[cat] = frozenset(lexicon[cat])

    return lexicon

def loadWeightedLexiconFromSparse(filename):
    """Loads the perma lexicon from a sparse formatting word[, word], category
    returns a dictionary of frozensets"""
    lexFile = open(filename)
    comma = re.compile(r'\,\s*')
    cats = [] 
    lexicon = {}
    for line in lexFile:
        items = comma.split(line.rstrip())
        terms = items[0:-2]
        cat = items[-2]
        weight = items[-1]
        if cat not in lexicon:
            lexicon[cat] = dict()
        for term in terms:
            if term:
                lexicon[cat][term]=weight

    return lexicon


def loadLexiconFromDic(filename):
    """Loads a lexicon from a .dic file such as LIWC2001_English.dic"""
    lexicon = {}

    with open(filename, 'r') as f:
        reading_categories = False
        cat_map = {}

        for line in f.readlines():
            if line:
                line = line.strip('\r\n')
                line = line.strip()
                if not line:
                    continue
            if line[0] == '%':
                reading_categories = not reading_categories
                continue
            line_split = [l for l in line.split('\t') if l] 
            if reading_categories:
                cat_map[int(line_split[0])] = line_split[1].strip().upper()
                # the above looks like: {1:PRONOUN, 2:I, 3:WE, ...}
            else:
                word = line_split[0]
                if line_split[1][0] == '(':
                    print('warning: line [%s] partially discarded due to inconsistent formatting'%(line, ))
                    continue
                try:
                    word_categories = [cat_map[x] for x in list(map(int, line_split[1:]))]
                except KeyError as e:
                    print("Category", line_split.pop(line_split.index(str(e.args[0]))), "doesn't exist [word: %s]" % line_split[0])
                word_categories = [cat_map[x] for x in list(map(int, line_split[1:]))]
                for category in word_categories:
                    try:
                        lexicon[category].add(word)
                    except KeyError:
                        lexicon[category] = set()
                        lexicon[category].add(word)

    for category in list(lexicon.keys()):
        lexicon[category] = frozenset(lexicon[category])

    return lexicon


def loadLexiconFeatMapFromCSV(filename):
    """Load a lexicon from a csv"""
    import csv
    csvReader = csv.reader(open(filename, 'rUb'))
    header = next(csvReader) #should be topic, topic_label, ....
    
    lexicon = {}
    labels_used = {}
    for row in csvReader:
        topic, label = row[0:2]
        original_label = label.lower()
        new_label = None
        
        if original_label in labels_used:
            labels_used[original_label] += 1
            new_label = '%s_%d'%(original_label, labels_used[original_label] + 1)

        if new_label:
            lexicon[new_label] = set()
            lexicon[new_label].add(topic)
        else:
            lexicon[original_label] = set()
            lexicon[original_label].add(topic)
            
        labels_used[original_label] = labels_used.get(original_label, 1)

        print(topic, original_label, new_label)

    return lexicon

def loadLexiconFromTopicFile(filename):
    """Loads a lexicon from a topic file
    returns a dictionary of frozensets"""
    lexFile = open(filename, 'r')

    lexicon = {}
    for line in lexFile.readlines():
        line_split = line.split('\t')
        if len(line_split)>1:
            category = str(line_split[0])
            lexicon[category] = set()
            terms = line_split[2]
            term_split = terms.split(' ')
            for term in term_split:
                if term != '\n':
                    lexicon[category].add(term)

    for category in list(lexicon.keys()):
        lexicon[category] = frozenset(lexicon[category])

    return lexicon
   
def loadWeightedLexiconFromTopicFile(filename):
    """Loads a weighted lexicon 
    returns a dictionary of dictionaries"""
    lexFile = open(filename, 'r')

    lexicon = {}
    for line in lexFile.readlines():
        line_split = line.split('\t')
        if len(line_split)>1:
            category = str(line_split[0])
            weight = line_split[1]
            terms = line_split[2]
            lexicon[category] = {}
            term_split = terms.split(' ')
            for term in term_split:
                if term != '\n':
                    lexicon[category][term] = weight
    return lexicon
   
def loadWeightedLexiconFromTopicCSV(filename, threshold=None, keep_underscores=False):
    """Loads a weighted lexicon 
    returns a dictionary of dictionaries"""
    import csv
    csvReader = csv.reader(open(filename, 'r'))
    header = next(csvReader) #should be topic_id, word1, ...etc..
    print("Loading %s" % filename)
    lexicon = {}
    for row in csvReader:
        topic = row[0]
        wordScores = row[1:]
        words = wordScores[::2]
        weights = wordScores[1::2]
        weights = list(map(float, weights))

        if threshold:
            new_words = []
            new_weights = []
            keep_pairs_with_weights_above_this_number = weights[0] * threshold
            if threshold == float('-inf'):
                keep_pairs_with_weights_above_this_number = threshold
            for ii in range(len(weights)):
                if weights[ii] > keep_pairs_with_weights_above_this_number:
                    new_words.append(words[ii])
                    new_weights.append(weights[ii])
            words = new_words
            weights = new_weights

        #now the weight for word[i] is weight[i]
        lexicon[topic] = {}
        for ii in range(0, len(words)):
            word = words[ii]
            if not keep_underscores: word = word.replace('_', ' ')
            lexicon[topic][word] = weights[ii]
            #print >> sys.stderr, "topic: %s, word: %s, weight: %2.2f"%(topic, words[ii], weights[ii])
    print("Done, num_topics: %d" % len(lexicon))
    return lexicon

     
def interactiveGetSenses(cat, word):
    os.system('clear')
    print("\n[%s] \033[92m%s\033[0m\n%s" %(PERMA_CODES[cat].title(), word, '='*(len(word)+20)))
    print("\033[90m%s\033[0m\n"%PERMA_LONG_DEF[cat])
    currentSenses = set()
    POSs = {wn.NOUN: 'noun', wn.VERB: 'verb', wn.ADJ: 'adjective', wn.ADV: 'adverb'}
    for pos, posName in POSs.items():
        synsets = wn.synsets(word, pos)
        if synsets:
            print("\t%s:" % posName)
            i = 1
            wss = [None,]
            for syns in synsets:
                wss.append(syns.name+'.'+word)
                print("\t\t\033[92m %d: \033[0m(%s)\033[92m %s\033[0m" % (i, ', '.join([lemma.name for lemma in syns.lemmas]), syns.definition))
                i+=1
            answered = False
            senses = None
            while not senses: 
                
                print("\n\tWhich of the above senses expresses \033[1m%s (i.e. %s)\033[0m?" % (PERMA_CODES[cat].title(), PERMA_SHORT_DEF[cat]))
                senses = input("\t(separate with spaces; 0 => none; cntrl-c to quit)? ")

                #validity check:
                senses = senses.strip()
                if not re.match(r'^[0-9, ]+$', senses):
                    print("entered non-numeric character")
                    senses = None
                    continue
                senses = re.findall(r'\d+', senses)
                ins = set(range(len(wss)))
                for s in senses:
                    s = int(s)
                    if s == 0 and len(senses) > 1:
                        print("entered 0 along with other senses")
                        senses = None
                    if s not in ins:
                        print("%d not a choice" % s)
                        senses = None

                #add to set:
            for s in senses:
                if s > 0:
                    ws = wss[int(s)]
                    print("\t\t\tadding %s" % ws)
                    currentSenses.add(ws)

    print("The following will be added: %s" % currentSenses)
    return currentSenses
                


#################################################################
## CLASS SETUP
#
class Lexicon(object):

    #instance variables:
    dbConn = None
    dbCursor = None
    currentLexicon = None

    def __init__(self, lex = None, mysql_host = dlac.MYSQL_HOST, lexicon_db=dlac.DEF_LEXICON_DB):
        (self.dbConn, self.dbCursor, self.dictCursor) = dbConnect(db=lexicon_db, host=mysql_host)
        self.mysql_host = mysql_host
        self.lexicon_db = lexicon_db
        self.currentLexicon = lex

    def __str__(self):
        return str(self.currentLexicon)

    ############################################################
    ## Instance Methods
    #

    def createLexiconTable(self, tablename):
        """Creates a lexicon table from the instances lexicon variable"""
        
        #first create the table:
        enumCats = "'"+"', '".join([k.upper().replace("'", "\\'") for k in list(self.currentLexicon.keys())])+"'"   
        drop = """DROP TABLE IF EXISTS """+tablename
        sql = """CREATE TABLE IF NOT EXISTS %s (id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY,
                 term VARCHAR(128), category ENUM(%s), INDEX(term), INDEX(category)) CHARACTER SET %s COLLATE %s ENGINE=%s""" % (tablename, enumCats, dlac.DEF_ENCODING, dlac.DEF_COLLATIONS[dlac.DEF_ENCODING.lower()], dlac.DEF_MYSQL_ENGINE)
        print("Running: ", drop)
        print("and:     ", sql)
        try:
            self.dbCursor.execute(drop)
            self.dbCursor.execute(sql)
        except MySQLdb.Error as e:
            dlac.warn("MYSQL ERROR in createLexiconTable" + str(e))
            sys.exit(1)

        #next insert rows:
        self.insertLexiconRows(tablename)

    def insertLexiconRows(self, tablename, lex=None):
        """Adds rows, taken from the lexicon variable to mysql"""
        if not lex: lex = self.currentLexicon
        #SETUP QUERY:
        sqlQuery = """INSERT INTO """+tablename+""" (term, category) values (%s, %s)"""
        values = []
        for cat, terms in lex.items():
            values.extend([[term, cat.upper()] for term in terms])
        
        try:
            self.dbCursor.executemany(sqlQuery, values)
        except MySQLdb.Error as e:
            dlac.warn("MYSQL ERROR in insertLexiconRows:" + str(e) + sqlQuery);
            sys.exit(1)

            
    def setLexicon(self, lexicon):
        self.currentLexicon = lexicon
    def getLexicon(self, temp="nothing"):
        return self.currentLexicon

    def loadLexicon(self, tablename, where = ''):
        """Loads a lexicon as currentLexicon"""
        sqlQuery = "SELECT term, category from "+tablename
        if where:
            sqlQuery += ' WHERE ' + where
        
        data = []
#        try:
        self.dbCursor.execute(sqlQuery)
        data = self.dbCursor.fetchall()
#        except MySQLdb.Error, e:
#            dlac.warn(" MYSQL ERROR" + str(e))
#            sys.exit(1)
        lexicon = {}
        for row in data:
            if not row[1] in lexicon:
                lexicon[row[1]] = []
            lexicon[row[1]].append(row[0])

        self.currentLexicon = {}
        for cat, terms in lexicon.items():
            self.currentLexicon[cat] = frozenset(terms)

        return self.currentLexicon
    
    def depolCategories(self):
        newDict = {}
        myDict = self.currentLexicon

        for cat in list(myDict.keys()):
            newCat = cat.rstrip('+').rstrip('-')
            if not newCat in newDict:
                newDict[newCat] = []
            newDict[newCat].extend(myDict[cat])
            
        for newCat in list(newDict.keys()):
            newDict[newCat] = frozenset(newDict[newCat])

        newLexicon = Lexicon()
        newLexicon.setLexicon(newDict)
        return newLexicon

    def unGroupCategories(self):
        newDict = {}
        myDict = self.currentLexicon

        for cat, terms in myDict.items():
            for term in terms:
                newDict[term.replace(' ', '_')+"_"+cat] = [term]
            
        for newCat in list(newDict.keys()):
            newDict[newCat] = frozenset(newDict[newCat])

        newLexicon = Lexicon()
        newLexicon.setLexicon(newDict)
        return newLexicon

    def union(self, otherLexicon):
        """union self lexicon with another and returns the result"""
        newDict = {}
        otherDict = otherLexicon.currentLexicon
        myDict = self.currentLexicon
        #print myDict
        #print otherDict
        unionKeys = set(myDict.keys()).union(set(otherDict.keys()))
        for cat in unionKeys:
            if (cat in myDict) and (cat in otherDict):
                newDict[cat] = frozenset(set(myDict[cat]).union(set(otherDict[cat])))
            else:
                if (cat in myDict):
                    newDict[cat] = frozenset(myDict[cat])
                else:
                    newDict[cat] = frozenset(otherDict[cat])
                
        newLexicon = Lexicon()
        newLexicon.setLexicon(newDict)
        return newLexicon

    def intersect(self, otherLexicon):
        """intersects self lexicon with another and returns the result"""
        newDict = {}
        otherDict = otherLexicon.currentLexicon
        myDict = self.currentLexicon
        unionKeys = set(myDict.keys()).union(set(otherDict.keys()))
        for cat in unionKeys:
            if (cat in myDict) and (cat in otherDict):
                newDict[cat] = frozenset(set(myDict[cat]).intersection(set(otherDict[cat])))
                
        newLexicon = Lexicon()
        newLexicon.setLexicon(newDict)
        return newLexicon
        

    def randomize(self):
        """randomizes the categories of the current lexicon"""
        myDict = self.currentLexicon
        myKeys = list(myDict.keys())
        newDict = {}
        for terms in list(self.currentLexicon.values()):
            for term in terms:
                randKey = myKeys[random.randint(0,len(myKeys)-1)]
                if not(randKey in newDict):
                    newDict[randKey] = []
                newDict[randKey].append(term)
        for cat, terms in newDict.items():
            newDict[cat] = frozenset(terms)
        newLex = Lexicon(newDict)
        newLex.setLexicon(newDict)
        return newLex


    def addTermsToCorpus(self, corpdb, corptable, termfield, messagefield, messageidfield, fulltext = False):
        """find rows with terms from lexicon and insert them back in as annotated rows"""
        #TODO: num_words and num_matches is hard-coded
        termList = set()
        for terms in list(self.currentLexicon.values()):
            termList = termList.union(terms)
        termREs = {}
        #escapedTerm = re.escape(term).replace('\\*', '\w*') #handle punctuation / add wildcard
        termREs = dict((term, re.compile(r'\b(%s)\b' % re.escape(term).replace('\\*', '\w*'), re.I)) for term in termList)
        termLCs = dict((term, term.rstrip('*').lower()) for term in termList)

        (corpDb, corpCursor) = abstractDBConnect(db=corpdb, host=dlac.MYSQL_HOST, user=dlac.USER)
        writeCursor = corpDb.cursor()
        #get field list:
        sql = """SELECT column_name FROM information_schema.columns WHERE table_name='%s' and table_schema='%s'""" % (corptable, corpdb)
        print(sql)
        corpCursor.execute(sql)
        rows = corpCursor.fetchall()
        fields = [r[0] for r in rows]
        fieldIndex = dict((fields[i], i) for i in range(len(fields)))

        #Go through each message      
        try:
            sql = """SELECT * FROM %s GROUP BY %s""" % (corptable, messageidfield)
            print(sql)
            corpCursor.execute(sql)
        except MySQLdb.Error as e:
            dlac.warn("MYSQL ERROR in addTermsToCorpus:" + str(e) + sqlQuery);
            sys.exit(1)
        newRows = []
        row = corpCursor.fetchone()
        record = 0
        sqlQuery = """REPLACE INTO """+corptable+""" values ("""+""", """.join(list('%s' for f in fields))+""")"""
        while row:
            if (len(row) != len(fields)):
                dlac.warn("row not correct size: " + str(row))
                sys.exit(1)

            else:
                record+=1
                message = row[fieldIndex[messagefield]]

                #first set number of words
                num_words = 0
                if message: 
                    num_words = len(message.split())

                message_id = row[fieldIndex[messageidfield]]
                updateSql = """UPDATE %s set num_words = %s WHERE %s = '%s'""" % (corptable, str(num_words), messageidfield, str(message_id))
                try:
                        #newRows = map (lambda r: r[f] = r[f].replace("'",  for f in range(fields))
                    writeCursor.execute(updateSql)
                except Exception as e:
                    dlac.warn("Exception during mysql call:" + str(e) + ': ' + sqlQuery);
                    pprint.PrettyPrinter().pprint(newRows) #debug
                    sys.exit(1)
            
                #then add extra rows:
                for term in termList:
                    if message and termLCs[term] in message.lower():
                        num_matches = len(termREs[term].findall(message))
                        if num_matches > 0:
                            newRow = list(row)
                            newRow[fieldIndex[termfield]] = term
                            newRow[fieldIndex['num_words']] = num_words
                            newRow[fieldIndex['num_matches']] = num_matches
                            newRow[fieldIndex['id']] = str(row[fieldIndex[messageidfield]])+'.'+term
                            newRows.append(newRow)
                if record % dlac.MAX_WRITE_RECORDS == 0: 
                    print("\n writing new rows up to: %d " % record)
                    #write them back in:
                    try:
                        #newRows = map (lambda r: r[f] = r[f].replace("'",  for f in range(fields))
                        writeCursor.executemany(sqlQuery, newRows)
                        newRows = []
                    except Exception as e:
                        dlac.warn("Exception during mysql call:" + str(e) + ': ' + sqlQuery);
                        pprint.PrettyPrinter().pprint(newRows) #debug
                        sys.exit(1)

            row = corpCursor.fetchone()

        print("Writing remaining. %d total rows checked" % record)
        try:
            writeCursor.executemany(sqlQuery, newRows)
        except Exception as e:
            dlac.warn("Exception during mysql call:" + str(e) + ': ' + sqlQuery);
            sys.exit(1)


    def createLexiconFromCorpus(self, corpdb, corptable, messagefield, messageidfield, minwordfreq):
        """Creates a lexicon (all in one category) from a examining word frequencies in a corpus"""
        wordList = dict()
        
        (corpDb, corpCursor) = abstractDBConnect(db=corpdb, host=dlac.MYSQL_HOST, user=dlac.USER)

        #Go through each message      
        try:
            sql = """SELECT %s FROM %s""" % (messagefield, corptable)
            if messageidfield: 
                sql += """ GROUP BY %s""" %  messageidfield
            dlac.warn(sql+"\n")
            corpCursor.execute(sql)
        except MySQLdb.Error as e:
            dlac.warn("MYSQL ERROR1:" + str(e) + sqlQuery)
            sys.exit(1)
        row = corpCursor.fetchone()
        rowNum = 0
        while row:
            rowNum+=1
            if rowNum % 10000 == 0:
                dlac.warn("On Row: %s\n" % rowNum)
            message = row[0]
            words = message.split()
            for word in words:
                if not word in wordList:
                    wordList[word] = 0
                wordList[word] += 1
            row = corpCursor.fetchone()

        #now create a set based on the words that occur more than minwordfreq
        lex = set()
        for word, freq in wordList.items():
            if (freq >= minwordfreq):
                lex.add(word)
        
        self.currentLexicon = {'word': lex}

    def annotateSenses(self, currentName, newLexiconName):
        #prompts the user to annotate the current lexicon, continues annotating if it exists
        weighted = self.isTableLexiconWeighted(currentName)
        senseLexicon = WeightedLexicon()
        try: 
            senseLexicon.loadLexicon(newLexiconName)
        except MySQLdb.Error as e:
            print("in except")
            #couldn't load lexicon, create it but empty
            createLike = """CREATE TABLE %s LIKE %s""" % (newLexiconName, currentName)
            self.dbCursor.execute(createLike)
            senseLexicon.loadLexicon(newLexiconName)

        #get lexicon dicts:
        oldLexicon = self.currentLexicon
        newLexicon = senseLexicon.currentLexicon

        #find what's been done
        seenWords = set()
        for cat, words in newLexicon.items():
            cat = cat.lower()
            for ws in words:
                if ws:
                    (lemma, pos, sense, word) = ws.split('.')
                    seenWords.add(cat+'#'+word)

        #prompt for new words
        for cat, words in oldLexicon.items():
            cat = cat.lower()
            for word in words:
                if cat+'#'+word in seenWords:
                    print("already annotated %s: %s (skipping)" % (cat, word))
                else:
                    senses = interactiveGetSenses(cat, word)
                    if senses:
                        if weighted:
                            smallNewLex = {cat: dict(list(zip(senses, [words[word]]*len(senses))))}
                            sys.stderr.write("newLexiconName %s \n" % newLexiconName)
                            self.insertWeightedLexiconRows(newLexiconName, smallNewLex)
                        else: 
                            smallNewLex = {cat: frozenset(senses)}
                            self.insertLexiconRows(newLexiconName, smallNewLex)
                    
    def expand(self):
        """Expands a lexicon to contain more words"""
        newLexicon = dict()
        for cat, words in self.currentLexicon.items():
            newLexicon[cat] = set()
            for word in words:
                print(word) #debug
                otherWords = Lexicon.wordExpand(word)
                newLexicon[cat] = newLexicon[cat].union(otherWords)
        
        return Lexicon(newLexicon)

    wpsRE = re.compile(r'[nvar]\.(\d+|\?)$')
    wpRE = re.compile(r'^([^\.#]+)\.([a-z])', re.I)

    @staticmethod
    def wordExpand(word,  specificDepth = 1, generalizeDepth = -1, totalLinks = 2):
        #specificDepth = 2 #length to travel toward more specific words
        #generalizeDepth = -1 #length (negative) to travel toward more general words
        # totalLinks: max links to travel along
        word = word.replace('#', '.')
        wpss = []
        words = set()
        if not Lexicon.wpsRE.search(word):
            match = Lexicon.wpRE.match(word)
            if match:
                pos = match.group(2)
                wpss = [(w, 0, 0) for w in wn.synsets(match.group(1)+'.'+pos)]
                words.add(match.group(1))
            else:
                wpss = [(w, 0, 0) for w in wn.synsets(word)]
                words.add(word)
        else:
            wpss = [(wn.synset(word), 0, 0)]

        seen = set()
        while wpss:
            wps, depth, links = wpss.pop()
            if wps not in seen:

                #add word to list:
                lemmas = [wps]
                if not isinstance(wps, str):
                    lemmas = [w.name for w in wps.lemmas]
                for lemma in lemmas:
                    wordMatch = Lexicon.wpRE.match(lemma)
                    if wordMatch:
                        words.add(wordMatch.group(1))
                    else:
                        words.add(lemma)

                #find neighbors:
                if links < totalLinks:
                    if depth < specificDepth:
                        #add hyponyms:
                        hypos = set(wps.hyponyms()  + 
                                     wps.entailments())
                        for hypo in hypos:
                            #print "%s hypo: %s" % (wps, hypo)
                            wpss.append((hypo, depth+1, links+1))
                    if depth > generalizeDepth:
                        #add hypernyms:
                        hypers = set(wps.hypernyms() + wps.instance_hypernyms() +
                                    wps.verb_groups())
                        for hype in hypers:
                            #print "%s hyper: %s" % (wps, hype)
                            wpss.append((hype, depth-1, links+1))

                seen.add(wps)

        #print "%s: %s" % (word, str(words))
        words = set([w.replace('_', ' ') for w in words])
        return words

    def likeExamples(self, corpdb, corptable, messagefield, numForEach = 60, onlyPrintIfMin = True, onlyPrintStartingAlpha = True):
        (corpDb, corpCursor) = abstractDBConnect(db=corpdb, host=dlac.MYSQL_HOST, user=dlac.USER)

        print("<html><head>")
        print("<style>")
        print("br {line-height: 6pt;}")
        print("li {margin:0; padding:0; margin-bottom:10pt;}")
        print("</style></head><body><table>")


        #csvFile = open('/tmp/examples.csv', 'w')

        for cat, terms in self.currentLexicon.items():
            print("<tr><td colspan=3> </td></tr>")
            print("<tr><td colspan=3><h3>%s</h3></td></tr>" % (cat))

            for term in sorted(terms):
                if term[0].isalpha() or not onlyPrintStartingAlpha:
                    escTerm = re.escape(term).replace('\\*', '\w*').replace('\\ ', '[[:space:]]').replace('(', '\(').replace(')', '\)') #handle punctuation / add wildcard
                    sql = """SELECT %s FROM %s WHERE %s RLIKE '[[:<:]]%s[[:>:]]'""" % (messagefield, corptable, messagefield, escTerm);
                    #print sql #debug
                    corpCursor.execute(sql)
                    messages = [x[0] for x in corpCursor.fetchall()]
                    lenmsgs = len(messages)
                    if not onlyPrintIfMin or lenmsgs >= numForEach:
                        print("<tr><td><b>%s</b></td><td><b>%s</b></td><td><em>%d occurrences</em></td></tr>" % (cat, term, lenmsgs))
                        print("<tr><td><b>%s</b></td><td><b>%s</b></td><td><b>%s</b></td></tr>" % ("Correct?", "term", "comment"))

                        toPrint = [] #holds messages to print
                        if lenmsgs < numForEach:
                            toPrint = messages
                        else:
                            done = set()
                            while len(toPrint) < numForEach:
                                i = random.randint(0, lenmsgs -1)
                                if i not in done:
                                    toPrint.append(messages[i])
                                    done.add(i)

                        tre = re.compile("\\b%s\\b"%(term), re.I)
                        for m in toPrint:
                            print("<tr><td></td><td>%s</td><td>%s</td></tr>" % (term, tre.sub("<b>%s</b>"%term, m)))
                        print("<tr><td colspan=3> </td></tr>")
        
        print("</table></body></html>")

    def likeSamples(self, corpdb, corptable, messagefield, category, lexicon_name, number_of_messages):
        (corpDb, corpCursor) = abstractDBConnect(db=corpdb, host=dlac.MYSQL_HOST, user=dlac.USER)

        #csvFile = open('/tmp/examples.csv', 'w')
        csvFile = open(lexicon_name+"_"+category+'.csv','wb')
#        csvFile.write('"id", "message", "term"\n')
        messages = list()
        def findTerm(xxx_todo_changeme):
            (m_id, string, term) = xxx_todo_changeme
            if 'space' in term:
                term = re.sub(r"\[\[:space:\]\]",r"\s",term, re.I)
            match = re.findall(term, string, re.I)
            return (m_id,string,match[0])
        for term in sorted(self.currentLexicon[category]):
            escTerm = re.escape(term).replace('\\*', '\w*').replace('\\ ', '[[:space:]]+') #handle punctuation / add wildcard
            sql = """SELECT id, %s FROM %s WHERE %s RLIKE '[[:<:]]%s[[:>:]]'""" % (messagefield, corptable, messagefield, escTerm);
                #print sql #debug
            corpCursor.execute(sql)
            print("Looking for messages containing %s" %  escTerm)
            new_ones = [(x[0],x[1],escTerm) for x in corpCursor.fetchall()]
            new_ones = list(map(findTerm,new_ones))
            messages.extend(new_ones)
        
        writer=csv.writer(csvFile,dialect='excel')
        random.shuffle(messages)
        writer.writerows(messages[:number_of_messages])
        

    def printCSV(self):
        """prints a csv style output of the lexicon"""
        ##print headers:
        print(','.join(iter(self.currentLexicon.keys())))

        ##turn sets into lists
        lexList = {}
        longest = 0;
        for cat, terms in self.currentLexicon.items():
            lexList[cat] = list(terms)
            longest = max(longest, len(terms))
        
        for i in range(longest):
            row = []
            for cat in self.currentLexicon.keys():
                if (i < len(lexList[cat])):
                    row.append(lexList[cat][i])
                else:
                    row.append('')
            print(','.join(row))

    # def printWeightedCSV(self):
    #     """prints a csv style output of the lexicon"""
    #     ##print headers:
    #     for cat, values in self.weightedLexicon.iteritems():
            
    #         topWords = sorted(values.items(), key=lambda x: x[1], reverse=True)[:30]
    #         topWString = ",",join(["'"+k+"',"+str(v) for k,v in topWords])
    #         print cat+','+topWString

            
    def pprint(self):
        """Uses pprint to print the current lexicon"""
        if self.currentLexicon:
            pprint.PrettyPrinter().pprint(self.currentLexicon)
        else:
            pprint.PrettyPrinter().pprint(self.weightedLexicon)

    def compare(self, otherLex):
        """Compares two lexicons, depends on the types"""
        uLexs = []
        wLexs = []
        for lex in (self, otherLex):
            if 'weightedLexicon' in lex.__dict__ and lex.weightedLexicon:
                wLexs.append(lex)
            else:
                uLexs.append(lex)

        if len(wLexs) > 1:
            return WeightedLexicon.compareWeightedToWeighted(self.weightedLexicon, otherLex.weightedLexicon)
        elif (len(wLexs) == 1):
            return WeightedLexicon.compareWeightedToUnweighted(wLexs[0].weightedLexicon, uLexs[0].currentLexicon)
        else:
            return WeightedLexicon.compareUnweightedToUnweighted(self.currentLexicon, otherLex.currentLexicon)


   

class WeightedLexicon(Lexicon):
    """WeightedLexicons have an additional dictionary with weights for each term in the regular lexicon"""
    def __init__(self, weightedLexicon=None, lex=None, mysql_host = dlac.MYSQL_HOST, lexicon_db=dlac.DEF_LEXICON_DB):
        super(WeightedLexicon, self).__init__(lex, mysql_host = mysql_host, lexicon_db = lexicon_db)
        self.weightedLexicon = weightedLexicon
     
    def isTableLexiconWeighted(self, tablename):
        sql = "SHOW COLUMNS from %s"%tablename
#        try:
        self.dbCursor.execute(sql)
        data = self.dbCursor.fetchall()
#        except MySQLdb.Error, e:
#            dlac.warn(" MYSQL ERROR" + str(e))
#            sys.exit(1)
        if len(data) > 0:
            numColumns = len(data)
            if numColumns == 3:
                return False
            elif numColumns == 4:
                return True
            else:
                raise Exception("Incorrect lexicon specified; number of rows in table [%s] is not 3 or 4")
        else:
            raise Exception("Lexicon table [%s] has no columns"%tablename)
            
        
    def isSelfLexiconWeighted(self):
        if self.weightedLexicon and isinstance(self.weightedLexicon, dict):
            for words in self.weightedLexicon.values():
                if isinstance(words, dict):
                    return True
                else:
                    return False
        else:
            return False
            
    def loadLexicon(self, tablename, where=''):
        """Loads a lexicon, checking to see if it is weighted or not, then responding accordingly"""
        if self.isTableLexiconWeighted(tablename):
            return self.loadWeightedLexicon(tablename, where)
        else:
            return super(WeightedLexicon, self).loadLexicon(tablename, where)
  
    def loadWeightedLexicon(self, tablename, where = ''):
        """Loads a lexicon as weightedLexicon"""
        sqlQuery = "SELECT term, category, weight from %s"%tablename
        if where:
            sqlQuery += ' WHERE ' + where
        
        data = []
#        try:
        self.dbCursor.execute(sqlQuery)
        data = self.dbCursor.fetchall()
#        except MySQLdb.Error, e:
#            dlac.warn(" MYSQL ERROR" + str(e))
#            sys.exit(1)
        lexicon = {}
        for (term, category, weight) in data:
            if not category in lexicon:
                lexicon[category] = {}
            lexicon[category][term] = weight

        self.weightedLexicon = lexicon
        return self.weightedLexicon
     
    def getWeightedLexicon(self):
        return self.weightedLexicon

    def createLexiconTable(self, tablename):
        """Loads a lexicon, checking to see if it is weighted or not, then responding accordingly"""
        if self.isSelfLexiconWeighted():
            print("Creating weighted lexicon table")
            return self.createWeightedLexiconTable(tablename)
        else:
            print("Creating unweighted lexicon table")
            return super(WeightedLexicon, self).createLexiconTable(tablename)

    def createWeightedLexiconTable(self, tablename):
        """Creates a lexicon table from the instance's lexicon variable"""
        
        #first create the table:
        enumCats = "'"+"', '".join([k.upper().replace("'", "\\'") for k in list(self.weightedLexicon.keys())])+"'"   
        drop = """DROP TABLE IF EXISTS """+tablename
        sql = """CREATE TABLE IF NOT EXISTS %s (id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, term VARCHAR(140), category ENUM(%s), weight DOUBLE, INDEX(term), INDEX(category)) CHARACTER SET %s COLLATE %s ENGINE=%s""" % (tablename, enumCats, dlac.DEF_ENCODING, dlac.DEF_COLLATIONS[dlac.DEF_ENCODING.lower()], dlac.DEF_MYSQL_ENGINE)
        print("Running: ", drop)
        print("and:     ", sql)
        try:
            self.dbCursor.execute(drop)
            self.dbCursor.execute(sql)
        except MySQLdb.Error as e:
            dlac.warn("MYSQL ERROR2" + str(e))
            sys.exit(1)

        #next insert rows:
        self.insertWeightedLexiconRows(tablename)
        print("Done creating lexicon: %s" % tablename)

    def createSuperTopicTable(self, tablename, reducedlex, lextable):
        """Creates a super topic table: unrolls a reduced lexicon at the word level"""
        
        if not lextable:
            print("You must supply a lex table")
            sys.exit(1)

        #first create the table:
        drop = """DROP TABLE IF EXISTS """+tablename
        sql = """CREATE TABLE {supertable} AS SELECT  orig.term, reduced.category, SUM(reduced.weight * orig.weight) AS weight 
                FROM {reducedtopic} reduced, {origtopic} orig 
                WHERE reduced.term=orig.category 
                GROUP BY reduced.category, orig.term""".format(supertable=tablename, reducedtopic=reducedlex, origtopic=lextable)
        id_col = """ALTER TABLE {supertable} ADD COLUMN `id` INT(11) NOT NULL AUTO_INCREMENT FIRST, ADD PRIMARY KEY (`id`);""".format(supertable=tablename)
        keys = """ALTER TABLE {supertable} ADD INDEX `term` (`term` ASC), ADD INDEX `category` (`category` ASC)""".format(supertable=tablename)
        
        try:
            print("Running: ", drop)
            self.dbCursor.execute(drop)
            print("and:     ", sql)
            self.dbCursor.execute(sql)
            print("and:     ", id_col)
            self.dbCursor.execute(id_col)
            print("Addings keys to term and category")
            self.dbCursor.execute(keys)
        except MySQLdb.Error as e:
            dlac.warn("MYSQL ERROR2" + str(e))
            sys.exit(1)

        print("Done creating super topics: %s" % tablename)

    def insertWeightedLexiconRows(self, tablename, lex = None):
        """Adds rows, taken from the lexicon variable to mysql"""
        if not lex: lex = self.weightedLexicon
        sqlQuery = """INSERT INTO """+tablename+""" (term, category, weight) values (%s, %s, %s)"""
        values = []
        for cat in lex:
            for term in lex[cat]:
                # print 'cat: %s term: %s' % (cat, term)
                if self.weightedLexicon[cat][term] != 0:
                    values.extend([[term, cat.upper(), self.weightedLexicon[cat][term]]])
        try:
            nbInserted = 0
            length = len(values)
            chunks = zip(*[iter(values)]*100)
            pprint.pprint(chunks)
            for v in chunks:
                nbInserted += self.dbCursor.executemany(sqlQuery, v)    
            remainingValues = values[nbInserted:]
            if remainingValues:
                nbInserted += self.dbCursor.executemany(sqlQuery, remainingValues)
            print("Inserted %d terms into the lexicon" % nbInserted)
            if nbInserted != length:
                print("Warning the number of rows inserted doesn't match the total number of rows")
                

        except MySQLdb.Error as e:
            dlac.warn("MYSQL ERROR:" + str(e) + sqlQuery);
            sys.exit(1)

    def mapToSuperLexicon(self, superLexiconMapping):
        """Creates a new lexicon based on mapping topic words to super topics """
        lex = self.weightedLexicon #category->word->weight ; lex[cat][word]
        mapping = superLexiconMapping.weightedLexicon
        superTopicDict = dict() #[word] -> [super_topic]-> [combined_weight]
        for superTopic, topicDict in mapping.items(): #super-topics
            superTopicDict[superTopic] = dict()
            for topic, topicWeight in topicDict.items():
                if topic in lex:
                    for word, wordWeight in lex[topic].items():
                        try: 
                            superTopicDict[superTopic][word] += wordWeight*topicWeight
                        except KeyError:
                            superTopicDict[superTopic][word] = wordWeight*topicWeight
                else:
                    print("Warning topic %s in super-topic but not original topic lexicon" % str(topic))

        return WeightedLexicon(weightedLexicon=superTopicDict)
            
    def union(self, otherLexicon):
        """union self lexicon with another and returns the result"""
        if not self.isSelfLexiconWeighted():
            print("union: not weighted")
            return super(WeightedLexicon, self).union(otherLexicon)
        newDict = {}
        otherDict = otherLexicon.weightedLexicon
        myDict = self.weightedLexicon
        unionKeys = set(myDict.keys()).union(set(otherDict.keys()))
        for cat in unionKeys:
            if (cat in myDict) and (cat in otherDict):
                newDict[cat] = dict(list(myDict.items()) + list(otherDict.items()))
            else:
                if (cat in myDict):
                    newDict[cat] = myDict[cat]
                else:
                    newDict[cat] = otherDict[cat]
                
        newLexicon = WeightedLexicon(newDict)
        return newLexicon

    def compare(self, otherLex):
        """Compares two lexicons, depends on the types"""
        uLexs = []
        wLexs = []
        for lex in (self, otherLex):
            if 'weightedLexicon' in lex.__dict__ and lex.weightedLexicon:
                wLexs.append(lex)
            else:
                uLexs.append(lex)

        if len(wLexs) > 1:
            return WeightedLexicon.compareWeightedToWeighted(self.weightedLexicon, otherLex.weightedLexicon)
        elif (len(wLexs) == 1):
            return WeightedLexicon.compareWeightedToUnweighted(wLexs[0].weightedLexicon, uLexs[0].currentLexicon)
        else:
            return WeightedLexicon.compareUnweightedToUnweighted(self.currentLexicon, otherLex.currentLexicon)


    def printCSV(self):
        """prints a csv style output of the lexicon"""
        if not self.isSelfLexiconWeighted():
            super(WeightedLexicon, self).printCSV()

        else:
            ##print headers:
            print("category, term1, w1, term2, w2, ...")

            ##print categories
            for cat, terms in self.weightedLexicon.items():
                print(cat+','+','.join(["%s,%d"%(term, w) for term, w in sorted(iter(terms.items()), key = lambda x: x[1], reverse=True)[:20]]))
                


    @staticmethod
    def compareWeightedToWeighted(wLex1, wLex2):
        """Compares two weighted lexicons"""
        #both wLex1 and wLex2 are dict of dicts: [category][word] = weight
        #TODO: Achal
        print("comparing weighted to weighted") #debug
        similarity = dict() #[cat1][cat2] = similarity_value
        for cat1 in wLex1.keys():
            similarity[cat1] = dict()
            for cat2 in wLex2.keys():
                similarity[cat1][cat2] = 0 #TODO

        return similarity

    @staticmethod
    def compareWeightedToUnweighted(wLex1, uLex2, comparisonMethod='weighted'):
        """Compares a weighted lexicon (self) to an unweighted lexicon"""
        #both wLex1 is a dict of dicts: [category][word] = weight
        #uLex2 is a dict of sets: [category]=set(<terms in category>)
        #TODO: Achal
        # shoudl either convert uLex2 to wLex2 and run compareWeighted.. or
        # convert wLex1 to uLex1 (by thresholding) and run compareUnweighted
        #(may not even use this code below)
        print("comparing weighted to unweighted") #debug
        similarity = dict() #[cat1][cat2] = similarity_value
        for cat1 in wLex1.keys():
            similarity[cat1] = dict()
            for cat2 in uLex2.keys():
                similarity[cat1][cat2] = 0 #TODO

        return similarity

    @staticmethod
    def compareUnweightedToUnweighted(uLex1, uLex2, metric='jaccard'):
        """Compares two unweighted lexicons"""
        #both uLex1 and uLex2 are dict of sets
        #TODO: Achal
        #run either jaccard or cosine sim (with probs equal to 1)
        print("comparing unweighted to unweighted") #debug
        similarity = dict() #[cat1][cat2] = similarity_value
        for cat1 in uLex1.keys():
            similarity[cat1] = dict()
            for cat2 in uLex2.keys():
                similarity[cat1][cat2] = 0 #TODO

        return similarity

class LexInterfaceParser(ArgumentParser):

    def __init__(self, description="On Features Class.", prefix_chars='-+', formatter_class=ArgumentDefaultsHelpFormatter, parents=None):
        ## Argument Parser ##

        super(LexInterfaceParser,self).__init__(description=description, prefix_chars=prefix_chars, formatter_class=formatter_class)

        group = self.add_argument_group('', '')
        group.add_argument("-f", "--file", dest="filename",
                              help="Lexicon Filename")
        group.add_argument("-g", "--gfile", dest="gfile",
                              help="Lexicon Filename in google format")
        group.add_argument("--sparsefile", dest="sparsefile",
                              help="Lexicon Filename in sparse format")
        group.add_argument("--weightedsparsefile", dest="weightedsparsefile",
                              help="Lexicon Filename in weighted sparse format")
        group.add_argument("--dicfile", dest="dicfile",
                              help="Lexicon Filename in dic (LIWC) format")
        group.add_argument("--topicfile", dest="topicfile",
                              help="Lexicon Filename in topic format")
        group.add_argument("--topic_csv", "--weighted_file", action='store_true', dest="topiccsv", default=False,
                              help="tells interface to use the topic csv format to make a weighted lexicon")
        group.add_argument("--keep_underscores", action='store_true', dest="keepunderscores", default=False,
                              help="Keep underscores in features in lexicon. Used with --topic_csv")
        group.add_argument("--filter", action="store_true", dest="using_filter",
                              help="Allows lexicon filtering if True")
        group.add_argument("-n", "--name", dest="name",
                              help="Existing Lexicon Table Name (will load)")
        group.add_argument("-c", "--create", dest="create",
                              help="Create a new lexicon table (must supply new lexicon name, and either -f, -g or -n)")
        group.add_argument("-p", "--print", action="store_true", dest="printcsv",
                              help="print lexicon to stdout (default csv format)")
        group.add_argument("--print_weighted", action="store_true", dest="printweightedcsv",
                              help="print lexicon to stdout (weighted csv format)")
        group.add_argument("--pprint", action="store_true", dest="pprint",
                              help="print lexicon to stdout as pprint output")
        group.add_argument("-w", "--where", dest="where",
                              help="where phrase to add to sql query")
        group.add_argument("-u", "--union", dest="union",
                              help="Unions two tables and uses the result as myLexicon")
        group.add_argument("-i", "--intersect", dest="intersect",
                              help="Intersects two tables and uses the result as myLexicon")
        group.add_argument("--super_topic", type=str, dest="supertopic",
                              help="Maps the current lexicon with a super topic mapping lexicon to make a super_topic"),
        group.add_argument("-r", "--randomize", action="store_true", dest="randomize",
                              help="Randomizes the categories of terms")
        group.add_argument("--depol", action="store_true", dest="depol",
                              help="Depolarize the categories (removes +/-)")
        group.add_argument("--ungroup", action="store_true", dest="ungroup", default = False,
                              help="places each word in its own category")
        group.add_argument("--compare", dest="compare",
                              help="Unions two tables and uses the result as myLexicon")
        group.add_argument("--annotate_senses", dest="sense_annotated_lex", type=str,
                              help="Asks the user to annotate senses of words and creates a new lexicon with senses (new lexicon name is the parameter)")

        group.add_argument("--topic_threshold", type=float, dest="topicthreshold", default=None,
                              help="sets the threshold to use for a csv topicfile")

        group.add_argument("-a", "--add_terms", action="store_true", dest="addterms",
                              help="Adds terms from the loaded lexicon to a given corpus (options below)")
        group.add_argument("-l", "--corpus_lexicon", action="store_true", dest="corpuslex",
                              help="Load a lexicon based on finding words in a given corpus (BETA) (options below)")
        group.add_argument("--corpus_examples", action="store_true", dest="examples",
                              help="Find example instances of words in the given corpus (using rlike; equal number for all words)")
        group.add_argument("--corpus_samples", action="store_true", dest="samples",
                              help="Find sample of matches for lexicon.")
        group.add_argument("-e", "--expand_lexicon", action="store_true", dest="expand",
                              help="Expands the lexicon to more terms.")

        group = self.add_argument_group('Terms OR Corpus Lexicon Options', '')
        group.add_argument("-d", "--corpus_db", dest="corpdb", metavar='DB', default = dlac.DEF_CORPDB,
                             help="Corpus database to use [default: %(default)s]")
        group.add_argument("-t", "--corpus_table", dest="corptable", metavar='TABLE', default = dlac.DEF_CORPTABLE,
                             help="Corpus table to use [default: %(default)s]")
        group.add_argument('--lexicondb', metavar='DB', dest='lexicondb', default=dlac.DEF_LEXICON_DB,
                            help='The database which stores all lexicons.')
        group.add_argument("--corpus_term_field", dest="termfield", metavar='FIELD', default = dlac.DEF_TERM_FIELD    ,
                             help="field of the corpus table that contains terms (lexicon table always uses 'term') [default: %(default)s]")
        group.add_argument("--corpus_message_field", dest="messagefield", metavar='FIELD', default = dlac.DEF_MESSAGE_FIELD    ,
                             help="field of the corpus table that contains the actual message [default: %(default)s]")
        group.add_argument("--corpus_messageid_field", dest="messageidfield", metavar='FIELD', default = dlac.DEF_MESSAGEID_FIELD    ,
                             help="field of the table that contains message ids (set to '' to not use group by [default: %(default)s]")
        group.add_argument("--min_word_freq", dest="minwordfreq", metavar='NUM', type=int, default = dlac.DEF_MIN_WORD_FREQ    ,
                             help="minimum number of instances to include in lexicon (-l option) [default: %(default)d]")
        group.add_argument("--lexicon_category", dest="lexicon_cat", metavar="CATEGORY", 
                             help="category in lexicon to get random samples from")
        group.add_argument("--num_rand_messages", dest="num_messages", metavar="NUM", type=int, default = dlac.DEF_NUM_RAND_MESSAGES,
                             help="number of random messages to select when getting samples from lexicon category")
        group.add_argument("--lex_interface", dest="lex_interface", action="store_true", help="Dummy arg for parsing from dlatkInterface.py") 

        ## Initialize any Objects ##
        self.lexicon = None 



    def processLDAInterface(self, args):
        """Main argument processing area"""
        ##Add Argument Processing here
        
        if args.name:
            self.lexicon = WeightedLexicon(lexicon_db=args.lexicondb)
            self.lexicon.loadLexicon(args.name, args.where)
        if args.filename:
            self.lexicon = WeightedLexicon(lexicon_db=args.lexicondb)
            self.lexicon.setLexicon(loadLexiconFromFile(args.filename))
        if args.gfile:
            self.lexicon = WeightedLexicon(lexicon_db=args.lexicondb)
            self.lexicon.setLexicon(loadLexiconFromGFile(args.gfile, args.using_filter))
        if args.sparsefile:
            self.lexicon = WeightedLexicon(lexicon_db=args.lexicondb)
            self.lexicon.setLexicon(loadLexiconFromSparse(args.sparsefile))
        if args.dicfile:
            self.lexicon = WeightedLexicon(lexicon_db=args.lexicondb)
            self.lexicon.setLexicon(loadLexiconFromDic(args.dicfile))
        if args.weightedsparsefile:
            self.lexicon = WeightedLexicon(loadWeightedLexiconFromSparse(args.weightedsparsefile), lexicon_db=args.lexicondb)
        if args.topicfile:
            self.lexicon = WeightedLexicon(lexicon_db=args.lexicondb)
            if args.topiccsv:
                self.lexicon = WeightedLexicon(loadWeightedLexiconFromTopicCSV(args.topicfile, args.topicthreshold, keep_underscores=args.keepunderscores), lexicon_db=args.lexicondb)
            else:
                self.lexicon.setLexicon(loadLexiconFromTopicFile(args.topicfile))
        if args.corpuslex:
            self.lexicon = WeightedLexicon(lexicon_db=args.lexicondb)
            self.lexicon.createLexiconFromCorpus(args.corpdb, args.corptable, args.messagefield, args.messageidfield, args.minwordfreq)
        if args.union:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            otherLexicon = WeightedLexicon(lexicon_db=args.lexicondb)
            otherLexicon.loadLexicon(args.union)
            self.lexicon = self.lexicon.union(otherLexicon)
        if args.intersect:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            otherLexicon = WeightedLexicon(lexicon_db=args.lexicondb)
            otherLexicon.loadLexicon(args.intersect)
            self.lexicon = self.lexicon.intersect(otherLexicon)
        if args.supertopic:
            superLexiconMapping = WeightedLexicon(lexicon_db=args.lexicondb)
            superLexiconMapping.loadLexicon(args.supertopic)
            self.lexicon = self.lexicon.mapToSuperLexicon(superLexiconMapping)
        if args.randomize:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon = self.lexicon.randomize()
        if args.depol:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon = self.lexicon.depolCategories()
        if args.ungroup:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon = self.lexicon.unGroupCategories()
        if args.compare:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            otherLexicon = WeightedLexicon(lexicon_db=args.lexicondb)
            otherLexicon.loadLexicon(args.compare)
            pprint.PrettyPrinter().pprint(self.lexicon.compare(otherLexicon))

        if args.sense_annotated_lex:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon.annotateSenses(args.name, args.sense_annotated_lex)

        if args.expand:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon = self.lexicon.expand()

        if args.create:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon.createLexiconTable(args.create)
        if args.addterms:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon.addTermsToCorpus(args.corpdb, args.corptable, args.termfield, args.messagefield, args.messageidfield, args.fulltext)
        if args.examples:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon.likeExamples(args.corpdb, args.corptable, args.messagefield)
        if args.samples:
            if not args.lexicon_cat:
                print("Must specify a lexicon category with option '--lexicon_cat'")
                sys.exit()
            self.lexicon.likeSamples(args.corpdb, args.corptable, args.messagefield, args.lexicon_cat, args.name, args.num_messages)
        if args.printcsv:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon.printCSV()
        if args.printweightedcsv:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon.printWeightedCSV()
        if args.pprint:
            if not self.lexicon:
                print("Must load a lexicon, either from a file (-f), or from another table (-n)")
                sys.exit()
            self.lexicon.pprint()



    def getParser(self):
        """Just incase someone is confused by the inheritance"""
        return self

    ###### Shouldn't Need to Edit Below Here ########

    ### These methods can be overWritten when creating subclass ###

    def processLoad(self, args):
        """processing state load arguments"""
        #seperated to make things easier for other to inherit this module
        #(can be overridden)

        #load:
        if args.load:
            objTup = self.load(args.savedir, args.load)
            if isinstance(objTup, tuple):
                (self.ol, self.lf) = objTup
            else:
                print("Nothing to load")     

        if args.liststates:
            self.printstates(args.savedir)

        if args.drop:
            self.drop(args.drop)

    def processSave(self, args):
        """processes the save arguments"""
        #separated to make it easy for others to put last
        #(can be overridden)
        if args.printObj:
            self.printObjs()

        if args.save:
            self.save(args.savedir, args.save, (self.ol, self.lf))

    def processArgs(self, args = '', parents = False):
        """Processes all arguments"""

        ##LOAD ARGUMENTS##
        if parents or not args:
            args = self.parse_args()

        ##PROCESS ARGUMENTS##
        self.processLDAInterface(args)

    def printObjs(self):
        if self.ol and not self.lf:
            print("OntoNotes Data:")
            print(self.ol)
        if self.lf:
            print("Language Features:")
            print(self.lf)

    def drop(self, objNames):
        for name in objNames:
            self.__dict__[name] = None

            
    ##HELPER FUNCTIONS:
    saveExtension = 'pickle'
    @staticmethod
    def save(savedir, savename, objectTup):
        saveFile = "%s/%s.%s" % (savedir, savename, OnFeaturesParser.saveExtension)
        print("Saving state to: %s" % saveFile)
        pickle.dump( objectTup, open( saveFile, "wb" ) )

    @staticmethod
    def load(savedir, savename):
        saveFile = "%s/%s.%s" % (savedir, savename, OnFeaturesParser.saveExtension)
        print("Loading state from: %s" % saveFile)
        objectTup = pickle.load( open( saveFile, "rb" ) )
        return objectTup

    @staticmethod
    def removeStateFile(savedir, savename):
        saveFile = "%s/%s.%s" % (savedir, savename, OnFeaturesParser.saveExtension)
        print("Removing state file: %s" % saveFile)
        os.remove(saveFile)

    matchExtension = re.compile(r'^(.*)\.'+saveExtension+'$')
    @staticmethod
    def printstates(savedir):
        files = os.listdir(savedir)
        names = []
        for fname in files:
            mObj = OnFeaturesParser.matchExtension.match(fname)
            if mObj:
                if os.stat("%s/%s"%(savedir,fname)).st_size > 0:
                    names.append((mObj.group(1), int(os.stat("%s/%s"%(savedir,fname)).st_size /1048576) ))
        if names:
            print("\nThe following saved states are available for loading:\n")
            print("  %-36s %12s" % ('NAME', 'SIZE'))
            print("  %-36s %12s" % ('----', '----'))
            for tup in sorted(names, key=lambda t: t[0]):
                print("  %-36s %10dMB" % tup)
        else:
            print("\nNo saved states available in directory: %s" % savedir)


###########################################################################
###########################################################################
if __name__ == "__main__":
    parser = LexInterfaceParser()
    parser.processArgs()




import re
from collections import OrderedDict
from pprint import pprint
import pandas as pd
import sys
from dateutil.parser import parse as dtParse
import datetime

#math / stats:
from numpy import sqrt, array, std, mean, log2, log
import numpy as np
import math
import numbers
from operator import mul
from functools import reduce
from sklearn.preprocessing import StandardScaler
from scipy.sparse import csr_matrix
from scipy.interpolate import interp1d

#local / nlp
from .lib.happierfuntokenizing import Tokenizer #Potts tokenizer

from .featureGetter import FeatureGetter
from . import dlaConstants as dlac
from .mysqlmethods import mysqlMethods as mm
from .mysqlmethods import mysql_iter_funcs as mif

class FeatureRefiner(FeatureGetter):
    """Deals with the refinement of feature information already in a table (outputs to new table)

    Returns
    -------
    FeatureRefiner object
    """

    def makeTopicLabelMap(self, topiclexicon, numtopicwords=5, is_weighted_lexicon=False):
        featlabel_tablename = 'feat_to_label$%s$%d'%(topiclexicon, numtopicwords)

        pldb = self.lexicondb
        (plconn, plcur, plcurD) = mm.dbConnect(pldb, charset=self.encoding, use_unicode=self.use_unicode)
        sql = 'DROP TABLE IF EXISTS `%s`'%featlabel_tablename
        mm.execute(pldb, plcur, sql, charset=self.encoding, use_unicode=self.use_unicode)
        sql = 'CREATE TABLE `%s` (`id` int(16) unsigned NOT NULL AUTO_INCREMENT, `term` varchar(128) DEFAULT NULL, `category` varchar(64) DEFAULT NULL, PRIMARY KEY (`id`), KEY `term` (`term`), KEY `category` (`category`) ) CHARACTER SET %s COLLATE %s ENGINE=%s' % (featlabel_tablename, self.encoding, dlac.DEF_COLLATIONS[self.encoding.lower()], dlac.DEF_MYSQL_ENGINE)
        mm.execute(pldb, plcur, sql, charset=self.encoding, use_unicode=self.use_unicode)

        sql = 'SELECT DISTINCT category FROM %s'%topiclexicon
        categories = [x[0] for x in mm.executeGetList(pldb, plcur, sql)]
        label_list = []
        for category in categories:
            if is_weighted_lexicon:
                sql = 'SELECT term, weight from %s WHERE category = \'%s\''%(topiclexicon, category)
                rows = mm.executeGetList(pldb, plcur, sql, charset=self.encoding, use_unicode=self.use_unicode)
                top_n_rows = sorted(rows, key=lambda x:x[1], reverse=True)
                terms = [x[0] for x in top_n_rows]
                label = ' '.join(map(str, terms[0:numtopicwords]))
                escaped_label = MySQLdb.escape_string(label)
                sql = 'INSERT INTO `%s` (`term`, `category`) VALUES(\'%s\', \'%s\')'%(featlabel_tablename, category, escaped_label )
                mm.execute(pldb, plcur, sql, charset=self.encoding, use_unicode=self.use_unicode)
            else:
                sql = 'SELECT term from %s WHERE category = \'%s\''%(topiclexicon, category)
                terms = [x[0] for x in mm.executeGetList(pldb, plcur, sql, charset=self.encoding, use_unicode=self.use_unicode)]
                label = ' '.join(map(str, terms[0:numtopicwords]))
                escaped_label = MySQLdb.escape_string(label)
                sql = 'INSERT INTO `%s` (`term`, `category`) VALUES(\'%s\', \'%s\')'%(featlabel_tablename, category, escaped_label )
                mm.execute(pldb, plcur, sql, charset=self.encoding, use_unicode=self.use_unicode)

        return featlabel_tablename

    def createCombinedFeatureTable(self, featureName = None, featureTables = [], tableName = None):
        """Create a new feature table by combining others"""

        #get best feat column type:
        intGrabber = re.compile(r'\d+')
        longestInt = 12
        longestType = "VARCHAR(12)"
        valueType = 'INTEGER'
        for table in featureTables:
            columns = mm.getTableColumnNameTypes(self.corpdb, self.dbCursor, table)
            if not columns: raise ValueError("One of your feature tables probably doesn't exist")
            currentType = columns['feat']
            valueType = columns['value']
            currentInt = int(intGrabber.search(currentType).group())
            if currentInt > longestInt:
                longestInt = currentInt
                longestType = currentType

        #get transformation:
        toNum = None
        numMatch = re.search(r'16to(\d+)', featureTables[0])
        if numMatch:
            toNum = numMatch.group(1)
        valueFunc = None
        if toNum:
            for func in dlac.POSSIBLE_VALUE_FUNCS:
                if float(func(16)) == float(toNum):
                    valueFunc = func
                    break     
        pocc = None
        poccmatch = re.search(r'\$([_0-9]+)\s*$', featureTables[0])
        if poccmatch:
            pocc = poccmatch.group(1)
        
        #CREATE TABLE:
        if not featureName:
            featNameGrabber = re.compile(r'^feat\$([^\$]+)\$')
            names = []
            for table in featureTables:
                names.append(featNameGrabber.match(table).group(1))
            featureName = '_'.join(names)
        featureTableName = self.createFeatureTable(featureName, "VARCHAR(%d)"%longestInt, valueType, tableName, valueFunc, extension=pocc)
        # Maarten: todo: test if too long and don't disable keys
        mm.disableTableKeys(self.corpdb, self.dbCursor, featureTableName, charset=self.encoding, use_unicode=self.use_unicode)#for faster, when enough space for repair by sorting

        for fTable in featureTables:
            mm.execute(self.corpdb, self.dbCursor, "INSERT INTO %s (group_id, feat, value, group_norm) SELECT group_id, feat, value, group_norm from %s;" % (featureTableName, fTable), charset=self.encoding, use_unicode=self.use_unicode)
        
        mm.enableTableKeys(self.corpdb, self.dbCursor, featureTableName, charset=self.encoding, use_unicode=self.use_unicode)#for faster, when enough space for repair by sorting

        return featureTableName


    def createTableWithBinnedFeats(self, num_bins, group_id_range, valueFunc = lambda x:x, 
                                   gender=None, genderattack=False, reporting_percent=0.04, outcomeTable = dlac.DEF_OUTCOME_TABLE, skip_binning=False):
        featureTable = self.featureTable
        group_id_range = list(map(int, group_id_range))
        newTable = featureTable+'$'+str(num_bins)+'b_'+'_'.join(map(str,group_id_range))
        if skip_binning: return newTable

        sql = 'DROP TABLE IF EXISTS %s'%newTable
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
        sql = "CREATE TABLE %s like %s" % (newTable, featureTable)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
        mm.standardizeTable(self.corpdb, self.dbCursor, newTable, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)

        groupNs = mm.executeGetList(self.corpdb, self.dbCursor, 'SELECT group_id, N FROM %s GROUP BY group_id'%self.featureTable, charset=self.encoding, use_unicode=self.use_unicode)
        groupIdToN = dict(groupNs)
        total_freq = sum([x[1] for x in groupNs])
        bin_size = float(total_freq) / float(num_bins+2)

        num_groups = len(groupNs)
        reporting_int = dlac._getReportingInt(reporting_percent, num_groups)

        # figure out the bins, i.e. if group_id's 1,2,3 total value is greater than "bin_size" our first bin is 1_3.
        dlac.warn('determining the number of bins...')
        current_sum = 0
        current_lower_group = groupNs[0][0]

        current_upper_group = None
        next_group_is_lower_group = False
        bin_groups = OrderedDict()
        gg = 0
        for group, value in groupNs:
            if next_group_is_lower_group:
                current_lower_group = group
                next_group_is_lower_group = False
            current_sum += value
            current_upper_group = group
            if current_sum >= bin_size:
                current_sum = 0
                bin_groups[(current_lower_group, current_upper_group)]  = '_'.join(map(str,[current_lower_group, current_upper_group]))
                next_group_is_lower_group = True
            gg += 1
            dlac._report('group_id\'s', gg, reporting_int, num_groups)
        if current_sum >= 0:
            bin_groups[(current_lower_group, current_upper_group)]  = '_'.join(map(str,[current_lower_group, current_upper_group]))

        max_label_length = max(list(map(len, list(bin_groups.values()))))

        sql = 'ALTER TABLE %s MODIFY COLUMN group_id VARCHAR(%d)'%(newTable, max_label_length) #this action preserves the index
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
        sql = 'ALTER TABLE %s ADD COLUMN `bin_center` float(6) not null default -1.0'%(newTable)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
        sql = 'ALTER TABLE %s ADD COLUMN `bin_center_w` float(6) not null default -1.0'%(newTable)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
        sql = 'ALTER TABLE %s ADD COLUMN `bin_width` int(10) not null default -1'%(newTable)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
        mm.disableTableKeys(self.corpdb, self.dbCursor, newTable, charset=self.encoding, use_unicode=self.use_unicode)

        # for each newly denoted bin: e.g. 1_3, 4_5, 6_6, ... get the new feature value counts / group norms; insert them into the new table
        # e.g. 1 'hi' 5, 2 'hi' 10, 3 'hi' 30 ==> 1_3 'hi' 45  (of course include group_norm also)
        dlac.warn('aggreagating the newly binned feature values / group_norms into the new table...')
        isql = 'INSERT INTO %s (group_id, feat, value, group_norm, std_dev, N, bin_center, bin_center_w, bin_width) VALUES (%s)'%(newTable, '%s, %s, %s, %s, %s, %s, %s, %s, %s')
        #isql = 'INSERT INTO %s (group_id, feat, value, group_norm, N, bin_center, bin_width) VALUES (%s)'%(newTable, '%s, %s, %s, %s, %s, %s, %s')
        ii_bins = 0
        num_bins = len(list(bin_groups.keys()))
        reporting_int = dlac._getReportingInt(reporting_percent, num_bins)
        #_warn('#############BIN NUMBER############### [[%d]] #############'%len(bin_groups))
        for (lower_group, upper_group), label in bin_groups.items():
            bin_N_sum = 0
            bin_width = 0
            bin_center = sum((lower_group, upper_group)) / 2.0
            bin_center_w = 0
            for ii in range(lower_group, upper_group+1):
                #_warn('for bin %d_%d ii:%d'%(lower_group, upper_group, ii))
                bin_width += 1
                bin_N_sum += groupIdToN.get(ii, 0)
                bin_center_w += groupIdToN.get(ii, 0) * ii
            bin_center_w = float(bin_center_w) / float(bin_N_sum)

            #_warn('number of users in range [%d, %d] is %d'%(lower_group, upper_group, bin_N_sum))
            
            # sql = 'SELECT group_id, feat, value, group_norm, N FROM %s where group_id >= %d AND group_id <= %d'%(self.featureTable, lower_group, upper_group)
            sql = 'SELECT group_id, feat, value, group_norm, std_dev FROM %s where group_id >= %d AND group_id <= %d'%(self.featureTable, lower_group, upper_group)
            groupFeatValueNorm = mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
            #pprint(groupFeatValueNorm)

            totalFeatCountForThisBin = float(0)
            featToValue = {}
            featToSummedNorm = {}
            for group_id, feat, value, norm, sd in groupFeatValueNorm:
            # for group_id, feat, value, norm, N in groupFeatValueNorm:
                if dlac.LOWERCASE_ONLY: feat = str(feat).lower()
                totalFeatCountForThisBin += value
                currentN = groupIdToN[group_id]
                try:
                    featToValue[feat] += value
                    featToSummedNorm[feat] += norm * currentN
                except KeyError:
                    featToValue[feat] = value
                    featToSummedNorm[feat] = norm * currentN

            #calculate mean and std_dev, using above info
            featToMeanNorm = {}
            featToSummedVar = {}
            for group_id, feat, _, norm, sd in groupFeatValueNorm:
                currentN = groupIdToN[group_id]
                meanNorm = featToSummedNorm[feat]/bin_N_sum
                try: 
                    featToSummedVar[feat] += currentN*((meanNorm - norm)**2 + (sd*sd))
                except KeyError:
                    featToSummedVar[feat] = currentN*((meanNorm - norm)**2 + (sd*sd))
                featToMeanNorm[feat] = meanNorm

            current_batch = [ ('_'.join(map(str,(lower_group, upper_group))),  k,  v, featToMeanNorm[k], sqrt(featToSummedVar[k] / bin_N_sum),
                               bin_N_sum, bin_center, bin_center_w, bin_width) for k, v in featToValue.items() ]
            mm.executeWriteMany(self.corpdb, self.dbCursor, isql, current_batch, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
            # print 'N bin sum:', bin_N_sum
            # isql = 'INSERT INTO %s (group_id, feat, value, group_norm, N, bin_center, bin_center_w, bin_width) VALUES (%s)'%(newTable, '%s, %s, %s, %s, %s, %s, %s, %s')
            ii_bins += 1
            dlac._report('group_id bins', ii_bins, reporting_int, num_bins)

        mm.enableTableKeys(self.corpdb, self.dbCursor, newTable, charset=self.encoding, use_unicode=self.use_unicode)
        dlac.warn('Done creating new group_id-binned feature table.')

        outputdata = mm.executeGetList(self.corpdb, self.dbCursor, 'select group_id, N from `%s` group by group_id'%(newTable,), charset=self.encoding, use_unicode=self.use_unicode)
        pprint(outputdata)

        # mm.execute(self.corpdb, self.dbCursor, 'drop table if exists `%s`'%(newTable,))
        return newTable

    def createTableWithRemovedFeats(self, p, minimumFeatSum = 0, groupFreqThresh = 0, setGFTWarning = False):
        """creates a new table with features that appear in more than p*|correl_field| rows, only considering groups above groupfreqthresh"""
        if not setGFTWarning:
            dlac.warn("""group_freq_thresh is set to %s. Be aware that groups might be removed during this process and your group norms will reflect this.""" % (groupFreqThresh), attention=True)
        toKeep = self._getKeepSet(p, minimumFeatSum, groupFreqThresh)
        label = str(p).replace('.', '_')
        if 'e' in label:
            label = label.replace('-','n')
        if minimumFeatSum > 1: 
            label += '_'+str(minimumFeatSum)
        return self.createNewTableWithGivenFeats(toKeep, label)

    def createNewTableWithGivenFeats(self, toKeep, label, featNorm=False):
        """Creates a new table only containing the given features"""

        featureTable = self.featureTable
        numToKeep = len(toKeep)
        newTable = featureTable+'$'+label
        mm.execute(self.corpdb, self.dbCursor, "DROP TABLE IF EXISTS %s" % newTable, charset=self.encoding, use_unicode=self.use_unicode)
        dlac.warn(" %s <new table %s will have %d distinct features.>" %(featureTable, newTable, numToKeep))
        sql = """CREATE TABLE %s like %s""" % (newTable, featureTable)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
        mm.standardizeTable(self.corpdb, self.dbCursor, newTable, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)
        mm.disableTableKeys(self.corpdb, self.dbCursor, newTable, charset=self.encoding, use_unicode=self.use_unicode)
  
        num_at_time = 2000
        total = 0
        toWrite = []
        
        wsql = """INSERT INTO """+newTable+""" (group_id, feat, value, group_norm, feat_norm) values (%s, %s, %s, %s, %s)""" if featNorm else """INSERT INTO """+newTable+""" (group_id, feat, value, group_norm) values (%s, %s, %s, %s)"""

        #iterate through each row, deciding whetehr to keep or not
        for featRow in self.getFeatAllSS(featNorm=featNorm):
            #print "%d %d" % (len(featRow), len(toWrite))
            if self.use_unicode and str(featRow[1]).lower() in toKeep:
                toWrite.append(featRow)
            elif not self.use_unicode and featRow[1].lower() in toKeep:
                toWrite.append(featRow)
            if len(toWrite) > num_at_time:
            #write those past the filter to the table
                mm.executeWriteMany(self.corpdb, self.dbCursor, wsql, toWrite, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
                total+= num_at_time
                if total % 100000 == 0: dlac.warn("%.1fm feature instances written" % (total/float(1000000)))
                toWrite = []

        #catch rest:
        if len(toWrite) > 0:
            #write those past the filter to the table
            mm.executeWriteMany(self.corpdb, self.dbCursor, wsql, toWrite, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)

        dlac.warn("Done inserting.\nEnabling keys.")
        mm.enableTableKeys(self.corpdb, self.dbCursor, newTable, charset=self.encoding, use_unicode=self.use_unicode)
        dlac.warn("done.")

        self.featureTable = newTable
        return newTable

    def _getKeepSet(self, p, minimumFeatSum = 0, groupFreqThresh = 0):
        """creates a set of features occuring in less than p*|correl_field| rows"""
        #acquire the number of groups (need to base on corp table):
        featureTable = self.featureTable
        totalGroups = self.countGroups(groupFreqThresh)
        assert totalGroups > 0, 'NO GROUPS TO FILTER BASED ON (LIKELY group_freq_thresh IS TOO HIGH)'
        assert p <= 1, 'p_occ > 1 not implemented yet'
        threshold = int(round(p*totalGroups))
        dlac.warn (" %s [threshold: %d]" %(featureTable, threshold))

        #acquire counts per feature (each row will come from a different correl_field)

        featCounts = self.getFeatureCounts(groupFreqThresh) #tuples of: feat, count (number of groups feature appears with)
        
        #apply filter:
        toKeep = set()
        i = 0
        for (feat, count) in featCounts:
            if count >= threshold:
                if self.use_unicode:
                    toKeep.add(str(feat).lower())
                else:
                    toKeep.add(feat.lower())
            i += 1
            
            if (i % 1000000) == 0: print("    checked %d features" % i)
        
        #apply secondary filter
        if minimumFeatSum > 1:
            featSums = self.getFeatureValueSums()
            for (feat, fsum) in featSums:
                if self.use_unicode:
                    feat = str(feat).lower()
                else:
                    feat = feat.lower()
                if feat in toKeep:
                    if fsum < minimumFeatSum:
                        toKeep.remove(feat)
            
        return toKeep

    def createStandardizedFeatTable(self, use_mean = True, use_std = True, groupFreqThresh = 0, setGFTWarning = False, where=None):
        #creates a copy of a feature table where all group_norms have been standardized
        #variables effecting standardization: GFT
        #use_mean : whether to mean center the number of words in the ngrams
        #use_std : whether to divide by standard deviation
        featTable = self.featureTable
        dlac.warn("""Standardizing %s \n -- note that if the table was sparse with many columns then this will make it very large unless use_mean is set to false.""" % (str(featTable)), attention=True)


        #1. apply GFT
        groups = []
        if not setGFTWarning:
            dlac.warn("""group_freq_thresh is set to %s. Be aware that groups might be used during mean and std calculation for standardizing.""" % (groupFreqThresh), attention=True)
        if groupFreqThresh:
            groupCnts = self.getGroupWordCounts(where)
            for group, wordCount in groupCnts.items():
                if (wordCount >= groupFreqThresh):
                    groups.append(group)
        else:
            groups = self.getDistinctGroups(where)

        #2. Get Sparse X:
        (groupNorms, featureNames) = self.getGroupNormsSparseFeatsFirst(groups)
        gnList = [groupNorms[feat] for feat in featureNames] #list of dictionaries of group_id => group_norm
        #featToIndex = dict([(featureNames[i], i) for i in range(len(featureNames))])
        groupToIndex = dict([(groups[i], i) for i in range(len(groups))])
        row = []
        col = []
        data = []
        # columns: features.
        for featIndex in range(len(gnList)):
            colData = gnList[featIndex]
            for groupid, value in colData.items():
                    row.append(groupToIndex[groupid])
                    col.append(featIndex)
                    data.append(value)
        assert all([isinstance(x,numbers.Number) for x in data]), "Data is corrupt, there are non float elements in the group norms (some might be NULL?)"
        X = csr_matrix((data,(row,col)), shape = (len(groups), len(featureNames)), dtype=np.float)
        if use_mean: 
            X = X.todense()
        dlac.warn("\n X.shape: %s]" % str(X.shape))

        #3. Apply STandardization: 
        scaler = StandardScaler(with_mean = use_mean, with_std=use_std)
        dlac.warn("\n [Applying StandardScaler to X: %s]" % str(scaler))
        X = scaler.fit_transform(X)
        dlac.warn(" [Done]\n")
        
        #4. Create new table:
        featTable = self.featureTable
        if isinstance(featTable,list):
            dlac.warn("Multiple  feature tables; only running on first: %s"%str(featTable))
            featTable = featTable[0]
        tableName = featTable+'$z'
        if not use_mean: tableName+='NoMe'
        if not use_std: tableName+='NoStd'
        columns = mm.getTableColumnNameTypes(self.corpdb, self.dbCursor, featTable)
        if not columns: raise ValueError("One of your feature tables probably doesn't exist")
        currentType = columns['feat']
        intGrabber = re.compile(r'\d+')
        featNameSize = int(intGrabber.search(currentType).group())
        featNameGrabber = re.compile(r'^feat\$([^\$]+)\$')
        featureType = featNameGrabber.match(featTable).group(1)
        
        tableName = self.createFeatureTable(featureType, "VARCHAR(%d)"%featNameSize, 'DOUBLE', tableName)


        #5. Insert into table
        if X.shape[0]*X.shape[1] < dlac.MAX_TO_DISABLE_KEYS: mm.disableTableKeys(self.corpdb, self.dbCursor, tableName, charset=self.encoding, use_unicode=self.use_unicode)#for faster, when enough space for repair by sorting
        for row in range(X.shape[0]):
            cf_id = groups[row]

            wsql = """INSERT INTO """+tableName+""" (group_id, feat, value, group_norm) values ('"""+str(cf_id)+"""', %s, %s, %s)"""
            if self.use_unicode:
                rows = [(featureNames[col], X[row][col], X[row][col]) for col in range(X.shape[1])]
            else:
                rows = [(featureNames[col].encode('utf-8'), X[row][col], X[row][col]) for col in range(X.shape[1])]
            mm.executeWriteMany(self.corpdb, self.dbCursor, wsql, rows, writeCursor=self.dbConn.cursor(), charset=self.encoding)

        dlac.warn("Done Reading / Inserting.")

        if X.shape[0]*X.shape[1] < dlac.MAX_TO_DISABLE_KEYS:
            dlac.warn("Adding Keys (if goes to keycache, then decrease MAX_TO_DISABLE_KEYS or run myisamchk -n).")
            mm.enableTableKeys(self.corpdb, self.dbCursor, featureTableName, charset=self.encoding, use_unicode=self.use_unicode)#rebuilds keys
        dlac.warn("Done\n")

        return tableName

    def createInterpolatedFeatTable(self, days = 1, dateField = 'created_at', minToImpute = 2, maxDaysUnitsEmpty = 4, groupFreqThresh = 0, setGFTWarning = False, where=None):
        #creates a new feature table at a higher level of aggregation
        #days: what units of time to interpolate into
        #maxDayUnitsEmpty: maximum number of days missed to keep a user.
        #TODO: handle maxDayUnitsEmpty

        featureTable = self.featureTable

        #0. Create new interpolated table:
        (_, name, corpTable, oldGroupField) = featureTable.split('$')[:4]
        assert oldGroupField == self.messageid_field, "Interpolate currently only works if interpolating message-level features"
        theRest = featureTable.split('$')[4:]
        feature_name = name[:11]+'_'+'_'.join([i[:3] for i in theRest])
        dayName = days
        if dayName%1.0 > 0: dayName = str(int(days*10)/10).replace('.', '_')
        else: dayName = str(int(dayName))
        newTable = 'feat$intrp'+ dayName \
                   + '_'+feature_name+'$'+corpTable+'$'+self.correl_field
        columns = mm.getTableColumnNameTypes(self.corpdb, self.dbCursor, featureTable)
        if not columns: raise ValueError("One of your feature tables probably doesn't exist")
        currentType = columns['feat']
        columns = mm.getTableColumnNameTypes(self.corpdb, self.dbCursor, self.corptable)
        if not columns: raise ValueError("One of your feature tables probably doesn't exist")
        groupidType = columns[self.correl_field]
        tableName = self.createFeatureTable(feature_name, currentType, 'DOUBLE', newTable)
        ##add extra columns:
        mm.execute(self.corpdb, self.dbCursor, "ALTER TABLE %s ADD time INT, ADD %s %s" % (tableName, self.correl_field, groupidType), charset=self.encoding, use_unicode=self.use_unicode)
     
        ## we are appending a date to the correl_field so group_id in feat table must be varchar
        if 'int' in groupidType:
            mm.execute(self.corpdb, self.dbCursor, "ALTER TABLE %s MODIFY group_id VARCHAR(64)" % (tableName), charset=self.encoding, use_unicode=self.use_unicode)
 
        dlac.warn("""Interpolating %s to the %s level.""" % (str(featureTable), self.correl_field))
        
        ## 1. apply GFT to get users we care about: 
        groups = []
        if not setGFTWarning:
            dlac.warn("""group_freq_thresh is set to %s. Only %s s meeting this will have interpolated features""" % (groupFreqThresh, self.correl_field), attention=True)
        if groupFreqThresh:
            groupCnts = self.getGroupWordCounts(where)
            for group, wordCount in groupCnts.items():
                if (wordCount >= groupFreqThresh):
                    groups.append(group)
        else:
            groups = self.getDistinctGroups(where)
        dlac.warn("""  Interpolating for up to %d %s s.""" % (len(groups), self.correl_field))
        gList = "','".join([str(g) for g in groups])

        ## 2. Get the minimum date:
        dlac.warn("""  [Finding good date range]""" )
        minIntersectDT, maxIntersectDT, minUnionDT, maxUnionDT  = mm.executeGetList(self.corpdb, self.dbCursor, "SELECT max(min_date), min(max_date), min(min_date), max(max_date) FROM (SELECT %s, MIN(%s) as min_date, MAX(%s) as max_date FROM %s where %s in ('%s') group by %s) as a " % \
                                         (self.correl_field, dateField, dateField, self.corptable, self.correl_field, gList, self.correl_field))[0]
        dtClosestToMin = mm.executeGetList(self.corpdb, self.dbCursor, "SELECT min(max_date) FROM (SELECT %s, MAX(%s) as max_date FROM %s where %s in ('%s') AND %s <= '%s' group by %s) as a " % \
                                         (self.correl_field, dateField, self.corptable, self.correl_field, gList, dateField, minIntersectDT, self.correl_field))[0][0]
        dtClosestToMax = mm.executeGetList(self.corpdb, self.dbCursor, "SELECT max(min_date) FROM (SELECT %s, MIN(%s) as min_date FROM %s where %s in ('%s') AND %s >= '%s' group by %s) as a " % \
                                         (self.correl_field, dateField, self.corptable, self.correl_field, gList, dateField, maxIntersectDT, self.correl_field))[0][0]
        if not isinstance(minIntersectDT, datetime.datetime):
            minIntersectDT = dtParse(minIntersectDT, ignoretz = True)
            maxIntersectDT = dtParse(maxIntersectDT, ignoretz = True)
            minUnionDT = dtParse(minUnionDT, ignoretz = True)
            maxUnionDT = dtParse(maxUnionDT, ignoretz = True)
            dtClosestToMin = dtParse(maxUnionDT, ignoretz = True)
            dtClosestToMax = dtParse(maxUnionDT, ignoretz = True)

            
            
        dayDiff = (maxIntersectDT - minIntersectDT).days
        maxDiffPerUnit = int(dayDiff/days)
        assert dayDiff > 0, "\nDayDiff (%d) was negative or zero. Increase GFT. Interpolating %ss over %ddays: 0 = %s through %d = %s"% \
                  (dayDiff, self.correl_field, days, str(minIntersectDT.date()), maxDiffPerUnit, str(maxIntersectDT.date()))

        ## 3. Get Features x SubIds (in Sparse X form):
        oldFeatures = FeatureGetter(self.corpdb, self.corptable, oldGroupField, self.mysql_host, self.message_field, self.messageid_field, self.encoding, True, self.lexicondb, featureTable)
        dateWhere = "%s >= DATE('%s') AND %s <= DATE('%s')" % (dateField, dtClosestToMin.date(), dateField, dtClosestToMax.date()) #used when querying mids + dates
        groupsWhere = "group_id in (SELECT %s FROM %s WHERE "%(oldGroupField, self.corptable) +dateWhere+" AND %s in ('%s'))" %(self.correl_field, gList)
        #print("groupsWhere", groupsWhere)#debug
        (groupNormsByMid, featureNames) = oldFeatures.getGroupNormsSparseGroupsFirst(where = groupsWhere)
        dlac.warn("""  [done]""")
        
        ## 4. Create X and Y per group: 
        lengroups = len(groups)
        dlac.warn("\nInterpolating %d %ss over %ddays: 0 = %s through %d = %s (full range: %s through %s)"% \
                  (lengroups, self.correl_field, days, str(minIntersectDT.date()), maxDiffPerUnit, \
                   str(maxIntersectDT.date()), str(dtClosestToMin.date()), str(dtClosestToMax.date())))
        gnum = 0
        wsql = """INSERT INTO """+newTable+""" (group_id, feat, value, group_norm, time, """+self.correl_field+""") values (%s, %s, %s, %s, %s, %s)"""
        for group in groups:

            #status tracker:
            if gnum%100==0:
                dlac.warn("   Read and considered interpolating %d %ss (%.2f complete) "%(gnum, self.correl_field, gnum/lengroups))
            gnum +=1

            ## 5. go throgh current users messages and align with feature table data:
            groupXYs = {f: [] for f in featureNames} #feat->[(X, y)...] where X is datediff from min and y is group_norm; adds zeros here
            midsAndDates = self.getMidAndExtraForCorrelField(group, dateField, where=dateWhere, warnMsg = False)
            if len(midsAndDates) < minToImpute:#make total messages is at least enough before bothering:
                dlac.warn(" !Only %d messages to impute for %s %s. Not enough. SKIPPING" % \
                          (len(midsAndDates), self.correl_field, str(group)))
                continue
            uniqueDtInts = set() #tracks total unique dts
            uniqueDtIntsInRange = set() #tracks total unique dts in the range
            for dateRow in midsAndDates:
                #print("dateRow", dateRow)#debug
                mid = dateRow[0]
                if mid in groupNormsByMid:
                    ## 6. get difference in dates by number of days parameter:
                    #print("checking", mid)
                    dt = dateRow[1]
                    if not isinstance(dt, datetime.datetime):
                        dt = dtParse(dt, ignoretz = True)
                    dayDiff = (dt - minIntersectDT).days
                    diffPerUnit = int(dayDiff/days)
                    uniqueDtInts.add(diffPerUnit)
                    if diffPerUnit >= 0:
                        uniqueDtIntsInRange.add(diffPerUnit)

                    for feat in featureNames:
                        try:
                            groupXYs[feat].append((diffPerUnit, groupNormsByMid[mid][feat]))
                        except KeyError: #feat must be missing in group Norms = 0
                            groupXYs[feat].append((diffPerUnit, 0))
            if len(uniqueDtInts) < minToImpute:#check that there are enough unique dates: 
                dlac.warn(" !Only %d unique dates to impute for %s %s (need at least %d). SKIPPING"% \
                          (len(uniqueDtInts), self.correl_field, str(group), minToImpute))
                continue
            elif len(uniqueDtIntsInRange) < minToImpute-1:#check that there are enough unique dates: 
                dlac.warn(" !Only %d unique dates i the range %s %s (need at least %d). SKIPPING"% \
                          (len(uniqueDtIntsInRange), self.correl_field, str(group), minToImpute-1))
                continue

            ## 7. Interpolate (note: this and 8 could be parallelized)
            minX, maxX = min(uniqueDtInts), max(uniqueDtInts)
            if maxX - minX < maxDiffPerUnit:#left and/or right will be left out
                dlac.warn(" !Warning, %s %s has smaller range (%d to %d) than max (0 to %d)."% \
                          (self.correl_field, str(group), minX, maxX, maxDiffPerUnit))
            newX = list(range(minX, maxX+1)) #i.e. range to interpolate over
            newYs = dict()
            newYs["_%ddInter"%days] = [1.0]*len(newX) #intercept to make sure day counts even if zero
            #fit:
            for feat in featureNames:
                x, y = zip(*groupXYs[feat])
                #print(x, y)
                try: 
                    f = interp1d(x, y, 'slinear')
                    newYs[feat] = f(newX)
                except ValueError:
                    dlac.warn("THIS SHOULD HAVE BEEN CAUGHT EARLIER.\n Not enough values to impute for %s %s. SKIPPING"%\
                              (self.correl_field, str(group), feat))
                    #NOTE: this should only happen on the first feat; because zeros have been added
                    break

            ## 8. Write to DB:
            #print(newX, newYs)#debug
            rows = []
            if not self.use_unicode:
                rows = [(str(group)+'_'+str(newX[i]), feat, newX[i], Ys[i], newX[i], group) for i in range(len(newX)) for feat, Ys in newYs.items() if Ys[i] != 0]
            else:
                rows = [((str(group)+'_'+str(newX[i])).encode('utf-8'), feat.encode('utf-8'), newX[i], \
                         Ys[i], newX[i], group) for i in range(len(newX)) for feat, Ys in newYs.items() if Ys[i] != 0]
            mm.executeWriteMany(self.corpdb, self.dbCursor, wsql, rows, writeCursor=self.dbConn.cursor(), charset=self.encoding)

            ##ADD USER_ID and DAYS so that can be used. 
            
        dlac.warn("Done Reading, Interpolating, and Inserting into '%s'."% tableName )
        mm.execute(self.corpdb, self.dbCursor, "ALTER TABLE %s ADD INDEX (time), ADD INDEX (%s)" % (tableName, self.correl_field), charset=self.encoding, use_unicode=self.use_unicode)

        return tableName


    
    def addFeatNorms(self, ReCompute = True, groupFreqThresh = 0, setGFTWarning=True):
        """Adds the mean normalization by feature (z-score) for each feature"""
        ##TODO: add feat_norm column if doesn't exist
        where = None
        if not ReCompute: where = 'feat_norm is null'
        groupNorms = self.getGroupNorms(where = where) #contains group_id, feat, group_norm
        if not setGFTWarning:
            dlac.warn("""group_freq_thresh is set to %s. Be aware that groups might be removed during this process and your group norms will reflect this.""" % (groupFreqThresh), attention=True)

        fMeans = self.addFeatTableMeans(groupNorms = groupNorms, groupFreqThresh=groupFreqThresh) #mean, std, zero

        wsql = """UPDATE """+self.featureTable+""" SET feat_norm = %s where group_id = %s AND feat = %s"""
        featNorms = []
        num_at_time = 2000
        numWritten = 0
        for (group_id, feat, group_norm) in groupNorms:
            if dlac.LOWERCASE_ONLY: feat = feat.lower()
            if (feat):
                fn = ( ((group_norm - fMeans[feat][0]) / float(fMeans[feat][1]), group_id, feat) )
                featNorms.append(fn)
                if len(featNorms) >= num_at_time:
                    mm.executeWriteMany(self.corpdb, self.dbCursor, wsql, featNorms, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
                    featNorms = []
                    numWritten += num_at_time
                    if numWritten % 100000 == 0: dlac.warn("%.1fm feature instances updated out of %dm" % 
                                                        ((numWritten/float(1000000)), len(groupNorms)/1000000))
                                    
        
        #write values back in 
        if featNorms: 
            mm.executeWriteMany(self.corpdb, self.dbCursor, wsql, featNorms, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)

        return True

        
    def findMeans(self, field='group_norm', addZeros = True, groupNorms = None, groups = set(), groupFreqThresh = 0, where = None):
        """Finds feature means from group norms"""

        if groupFreqThresh:
            groupCnts = self.getGroupWordCounts(where)
            for group, wordCount in groupCnts.items():
                if (wordCount >= groupFreqThresh):
                    groups.add(group)
        
        if not groupNorms:
            groupNorms = self.getGroupNorms() #contains group_id, feat, group_norm

        #turn into dict of lists:
        fNumDict = dict() 
        for (gid, feat, gn) in groupNorms:
            if (len(groups) < 1) or (gid in groups): 
                if dlac.LOWERCASE_ONLY: feat = feat.lower()
                if feat: 
                    try:
                        fNumDict[feat].append(gn)
                    except KeyError:
                        fNumDict[feat] = [gn]                      

        #add zeros for groups missing featrues
        if addZeros:
            totalGroups = self.countGroups(groupFreqThresh)
            for feat, nums in fNumDict.items():
                difference = totalGroups - len(nums)
                nums.extend([0]*difference)
            
        #calculate values:
        fMeans = dict()
        for feat, nums in fNumDict.items():
            nums = array([float(d) for d in nums])
            (m, s) = (mean(nums), std(nums))
            z = (0 - m) / float(s)
            fMeans[feat] = (m, s, z)

        return fMeans


    def addFeatTableMeans(self, field='group_norm', groupNorms = None, groupFreqThresh = 0, groups=set()):
        """Add to the feature mean table: mean, standard deviation, and zero_mean for the current feature table"""
       
        #CREATE TABLE
        meanTable = 'mean$'+self.featureTable
        mm.execute(self.corpdb, self.dbCursor, "DROP TABLE IF EXISTS %s" % meanTable, charset=self.encoding, use_unicode=self.use_unicode)
        featType = mm.executeGetList(self.corpdb, self.dbCursor, "SHOW COLUMNS FROM %s like 'feat'" % self.featureTable)[0][1]
        sql = """CREATE TABLE %s (feat %s, mean DOUBLE, std DOUBLE, zero_feat_norm DOUBLE, PRIMARY KEY (`feat`)) CHARACTER SET %s COLLATE %s ENGINE=%s""" % (meanTable, featType, self.encoding, dlac.DEF_COLLATIONS[self.encoding.lower()], dlac.DEF_MYSQL_ENGINE)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)

        fMeans = self.findMeans(field, True, groupNorms, groupFreqThresh = groupFreqThresh)
        fMeansList = [(k, v[0], v[1], v[2]) for k, v in fMeans.items()]
        #print fMeansList #debug

        #WRITE TO TABLE:
        sql = """INSERT INTO """+meanTable+""" (feat, mean, std, zero_feat_norm) VALUES (%s, %s, %s, %s)"""
        mm.executeWriteMany(self.corpdb, self.dbCursor, sql, fMeansList, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)

        return fMeans


    def createCollocRefinedFeatTable(self, threshold = 3.0, featNormTable=False):
        #n = the number of words in the ngrams
        #uses pmi to remove uncommon collocations:
        featureTable = self.featureTable
        dlac.warn(featureTable)
        wordGetter = self.getWordGetter()
        tokenizer = Tokenizer(use_unicode=self.use_unicode)

        jointFreqs = self.getSumValuesByFeat()
        wordFreqs = dict(wordGetter.getSumValuesByFeat())
        allFreqs = wordGetter.getSumValue()

        keepers = set()
        for (colloc, freq) in jointFreqs:
            # words = tokenizer.tokenize(colloc)
            # If words got truncated in the creation of 1grams, we need to account for that
            words = [word[:dlac.VARCHAR_WORD_LENGTH] for word in tokenizer.tokenize(colloc)]
            if (len(words) > 1):
                indFreqs = [wordFreqs[w] for w in words if w in wordFreqs]
                pmi = FeatureRefiner.pmi(freq, indFreqs, allFreqs, words = words)
                # print "%s: %.4f" % (colloc, pmi)#debug
                if pmi > (len(words)-1)*threshold: 
                    keepers.add(colloc)
            else:
                keepers.add(colloc)
        return self.createNewTableWithGivenFeats(keepers, "pmi%s"%str(threshold).replace('.', '_'), featNormTable)

    def getCollocsWithPMI(self):
        '''
        :inputs: self.featureTable
        calculates PMI for each ngram that is >1
        :returns: a dict of colloc => [pmi, num_tokens, pmi_threshold_val]
            **pmi_threshold_val is pmi/(num_tokens-1), thats what --feat_colloc_filter is based on
        '''
        featureTable = self.featureTable
        dlac.warn(featureTable)
        wordGetter = self.getWordGetter()
        tokenizer = Tokenizer(use_unicode=self.use_unicode)

        jointFreqs = self.getSumValuesByFeat()
        wordFreqs = dict(wordGetter.getSumValuesByFeat())
        allFreqs = wordGetter.getSumValue()

        keepers = set()
        collocPMIs = {}
        count = 0
        print("len(jointFreqs): " + str(len(jointFreqs)))
        for (colloc, freq) in jointFreqs:
            count +=1
            if count % 50000 == 0:
                print("calculating pmi for {}th feature".format(count))
            words = [word[:dlac.VARCHAR_WORD_LENGTH] for word in tokenizer.tokenize(colloc)]
            if (len(words) > 1):
                indFreqs = [wordFreqs[w] for w in words if w in wordFreqs]
                pmi = FeatureRefiner.pmi(freq, indFreqs, allFreqs, words = words)
                collocPMIs[colloc] =[colloc, freq, pmi, len(words), pmi/(len(words)-1)]
        return collocPMIs

    @staticmethod
    def pmi(jointFreq, indFreqs, allFreq, words = None):
        #  log p(w1, w2) / (p(w1)p(w2)) 
        allFreq = float(allFreq) #to insure floating point div.
        jointP = float(jointFreq) / allFreq
        
        denominator = 1.0
        for iFreq in indFreqs:
            denominator *= (float(iFreq) / allFreq)
        out = 0.0
        try:
            out = log2( jointP/denominator)
        except ZeroDivisionError:
            print(jointFreq, indFreqs, allFreq, words)
        return out

    @staticmethod
    def salience(jointFreq, indFreqs, allFreq):
        #  (log p(w1, w2) / (p(w1)p(w2)) * log f(w1, w2)
        return FeatureRefiner.pmi(jointFreq, indFreqs, allFreq) * log2(jointFreq)

    def createCorrelRefinedFeatTable(self, correls, pValue = 0.05, featNormTable=True):
        keepSet = set() #names of features to keep
        outcomes = set()
        for outcomeField, featCorrels in correls.items():
            outcomes.add(outcomeField)
            for feat, tup in featCorrels.items():
                (r, p, n) = tup[:3]
                if p <= pValue:
                    keepSet.add(feat)

        return self.createNewTableWithGivenFeats(keepSet, '_'.join([x[:3] for x in outcomes]), featNormTable)

    def getCorrelFieldType(self, correlField):
        if correlField == 'state':
            return 'char(2)'
        return None

    def createFeatureTable(self, featureName, featureType = 'VARCHAR(64)', valueType = 'INTEGER', tableName = None, valueFunc = None, correlField=None, extension = None):
        """Creates a feature table based on self data and feature name"""
        
        #create table name
        if not tableName: 
            valueExtension = ''
            tableName = 'feat$'+featureName+'$'+self.corptable+'$'+self.correl_field
            if valueFunc: 
                tableName += '$' + str(16)+'to'+"%d"%round(valueFunc(16))
            if extension: 
                tableName += '$' + extension

        #find correl_field type:
        sql = """SELECT column_type FROM information_schema.columns WHERE table_schema='%s' AND table_name='%s' AND column_name='%s'""" % (
            self.corpdb, self.corptable, self.correl_field)

        correlField = self.getCorrelFieldType(self.correl_field) if not correlField else correlField
        correl_fieldType = mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)[0][0] if not correlField else correlField

        #create sql
        drop = """DROP TABLE IF EXISTS %s""" % tableName
        # featureType = "VARCHAR(30)" # MAARTEN
        #CREATE TABLE feat_3gram_messages_rand1000_user_id (id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, user_id ('bigint(20) unsigned',), 3gram VARCHAR(64), VALUE INTEGER
        #sql = """CREATE TABLE %s (id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, group_id %s, feat %s, value %s, group_norm DOUBLE, feat_norm DOUBLE, KEY `correl_field` (`group_id`), KEY `feature` (`feat`)) CHARACTER SET utf8 COLLATE utf8_general_ci""" %(tableName, correl_fieldType, featureType, valueType)
        sql = """CREATE TABLE %s (id INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, group_id %s, feat %s, value %s, group_norm DOUBLE, KEY `correl_field` (`group_id`), KEY `feature` (`feat`)) CHARACTER SET %s COLLATE %s ENGINE=%s""" %(tableName, correl_fieldType, featureType, valueType, self.encoding, dlac.DEF_COLLATIONS[self.encoding.lower()], dlac.DEF_MYSQL_ENGINE)

        #run sql
        mm.execute(self.corpdb, self.dbCursor, drop, charset=self.encoding, use_unicode=self.use_unicode)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)

        return  tableName;


    def createFeatTableByDistinctOutcomes(self, outcomeGetter, controlValuesToAvg = [], outcomeRestriction = None, nameSuffix=None ):
        """Creates a new feature table, by combining values based on an outcome, then applies an averaging based on controls"""
        ##TODO: perform outcome restriction by using group freq thresh instead of uwt, for flexibility
        featureTable = self.featureTable
        outcomeTable = outcomeGetter.outcome_table
        assert len(outcomeGetter.outcome_value_fields) < 2, 'Currently, only allowed to specify one outcome.'
        outcomeField = outcomeGetter.outcome_value_fields[0]
        controlField = None
        if outcomeGetter.outcome_controls: 
            assert len(outcomeGetter.outcome_controls) < 2, 'Currently, only allowed to specify one control.'
            controlField = outcomeGetter.outcome_controls[0]
            if len(controlValuesToAvg) < 1:
                dlac.warn("getting distinct values for controls")
                controlValuesToAvg = outcomeGetter.getDistinctOutcomeValues(outcome = controlField, includeNull = False, where=outcomeRestriction)

        #create new table name:
        nameParts = featureTable.split('$')
        nameParts = [part.replace('16to', '') for part in nameParts]
        nameParts = [part.replace('messages', 'msgs') for part in nameParts]
        newTables = []
        nameSuffix = '' if not nameSuffix else '_%s'%(nameSuffix,)
        if controlField:
            for value in controlValuesToAvg:
                controlGroupName = outcomeField + '_' + controlField + '_' + str(value)
                newTables.append('feat_grpd'+ nameSuffix +'$' + '$'.join(nameParts[1:3]) + '$' + controlGroupName + '$' + '$'.join(nameParts[4:]))
        else: 
            newTables.append('feat_grpd'+ nameSuffix +'$' + '$'.join(nameParts[1:3]) + '$' + outcomeField + '$' + '$'.join(nameParts[4:]))

        #1. create table where outcome is group_id and insert values
        for newTable in newTables:
            drop = """DROP TABLE IF EXISTS %s""" % (newTable)
            sql = "create table %s like %s" % (newTable, featureTable)
            mm.execute(self.corpdb, self.dbCursor, drop, charset=self.encoding, use_unicode=self.use_unicode)
            mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
            sql = 'ALTER TABLE %s ADD COLUMN `N` int(16) not null default -1'%(newTable)
            mm.execute(self.corpdb, self.dbCursor, sql)
            sql = 'ALTER TABLE %s CHANGE feat_norm std_dev FLOAT' % newTable;
            mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
            mm.standardizeTable(self.corpdb, self.dbCursor, newTable, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)
            
        outres = outcomeRestriction
        outres = outres + ' AND ' if outres else '' #only need and if it exists
        if controlField:
            for outcomeValue, cntrlcounts in \
                    outcomeGetter.getDistinctOutcomeAndControlValueCounts(control = controlField, includeNull = False, where=outcomeRestriction).items():
                for cvalue, count in cntrlcounts.items():
                    if cvalue in controlValuesToAvg:
                        newTable = 'feat_grpd'+ nameSuffix + '$' + '$'.join(nameParts[1:3]) + '$' + outcomeField + '_' + controlField + '_' + str(cvalue) + '$' + '$'.join(nameParts[4:])
                        print("on %s %s and %s %s, count: %d" % (outcomeField, str(outcomeValue), controlField, str(cvalue), count))
                        sql = "INSERT INTO %s (group_id, feat, value, group_norm, std_dev, N) SELECT age, feat, total_freq, mean_rel_freq, SQRT((N_no_zero*(POW((mean_no_zero - mean_rel_freq), 2) + std_no_zero*std_no_zero) + (N - N_no_zero)*(mean_rel_freq * mean_rel_freq)) / N) as std, N  from (SELECT b.%s, feat, SUM(value) as total_freq, SUM(group_norm)/%d as mean_rel_freq, AVG(group_norm) as mean_no_zero, std(group_norm) as std_no_zero, %d as N, count(*) as N_no_zero FROM %s AS a, %s AS b WHERE %s b.%s = '%s' AND b.%s = '%s' AND b.user_id = a.group_id group by b.%s, a.feat) as stats" % (newTable, outcomeField, count, count, featureTable, outcomeTable, outres, controlField, str(cvalue), outcomeField, str(outcomeValue), outcomeField)
#SELECT age, feat, total_freq, mean_rel_freq, SQRT((N_no_zero*(POW((mean_no_zero - mean_rel_freq), 2) + std_no_zero*std_no_zero) + (N - N_no_zero)*(mean_rel_freq * mean_rel_freq)) / N) as std, N  from (
#SELECT b.age, feat, SUM(value) as total_freq, SUM(group_norm)/390 as mean_rel_freq, AVG(group_norm) as mean_no_zero, std(group_norm) as std_no_zero, 390 as N, count(*) as N_no_zero FROM feat$1gram$messages_en$user_id$16to16$0_01 AS a, masterstats_andy AS b WHERE UWT >= 1000 AND b.age = '45' AND b.user_id = a.group_id group by b.age, a.feat) as a             
                        mm.execute(self.corpdb, self.dbCursor, sql, False, charset=self.encoding, use_unicode=self.use_unicode)
                    else:
                        print("skipping %s %s and %s %s, count: %d because control value not in list" % (outcomeField, str(outcomeValue), controlField, str(cvalue), count))
        else: #no controls to avg
            # Maarten
            correspondences = outcomeGetter.getGroupAndOutcomeValues()
            correspondences_inv = {}
            for k,v in correspondences:
                correspondences_inv[v] = correspondences_inv.get(v,[])
                correspondences_inv[v].append(k)
            correspondences = correspondences_inv
            total_sum_values = {i[0]: int(i[1]) for i in self.getSumValuesByGroup()}

            i = 0
            j = 0
            for outcomeValue, groups in correspondences.items():
                i += 1
                rows = []
                groups_nonZero = [g for g in groups if g in total_sum_values]
                for feat, values, gns, Nfeats in self.yieldGroupNormsWithZerosByFeat(groups = groups, values = True):
                    if not values: continue

                    sum_value = sum(values.values())
                    total_sum_value = sum(total_sum_values[g] for g in groups_nonZero)
                    group_norm = float(sum_value)/total_sum_value
                    std_dev = std(list(gns.values()))
                    N = len(gns)
                    rows.append([outcomeValue, feat, sum_value, group_norm, std_dev, N])
                    if len(rows) >= dlac.MYSQL_BATCH_INSERT_SIZE:
                        sql = "INSERT INTO %s (group_id, feat, value, group_norm, std_dev, N) " % newTable
                        sql += "VALUES (%s)" % ', '.join('%s' for r in rows[0]) 
                        mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
                        j += len(rows)
                        print("    wrote %d rows [finished %d outcome_values]" % (j, i))
                        rows = []
                    
                if rows:
                    sql = "INSERT INTO %s (group_id, feat, value, group_norm, std_dev, N) " % newTable
                    sql += "VALUES (%s)" % ', '.join('%s' for r in rows[0])
                    mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
                    j += len(rows)  
                    print("    wrote %d rows [finished %d outcome_values]" % (j, i))
                print("Inserted into %s" % newTable)

                        
                """
                for outcomeValue, count in outcomeGetter.getDistinctOutcomeValueCounts(includeNull = False, where=outcomeRestriction).iteritems():
                
                newTable = 'feat_grpd'+ '$' + '$'.join(nameParts[1:3]) + '$' + outcomeField + '$' + '$'.join(nameParts[4:])
                print "on %s %s, count: %d (no control)" % (outcomeField, str(outcomeValue), count)
                sql = "INSERT INTO %s (group_id, feat, value, group_norm, std_dev, N) SELECT %s, feat, total_freq, mean_rel_freq, SQRT((N_no_zero*(POW((mean_no_zero - mean_rel_freq), 2) + std_no_zero*std_no_zero) + (N - N_no_zero)*(mean_rel_freq * mean_rel_freq)) / N) as std, N  from (SELECT group_id, feat, SUM(value) as total_freq, SUM(group_norm)/%d as mean_rel_freq, AVG(group_norm) as mean_no_zero, std(group_norm) as std_no_zero, count(1) as N_no_zero, %d as N FROM %s) AS a, %s AS b WHERE %s b.%s = '%s' AND b.%s = a.group_id group by b.%s, a.feat" % (newTable, outcomeField,  count, count, featureTable, outcomeTable, outres, outcomeField, str(outcomeValue), self.correl_field ,outcomeField)
                # print "Maarten", self.correl_field, sql
                mm.execute(self.corpdb, self.dbCursor, sql, False)"""
        #2: Combine feature table to take average of controls:
        #controlGroupAvgName = outcomeField + '_' + controlField + 'avg_' + '_'.join(map(lambda v: str(v), controlValuesToAvg))
        if controlField and len(newTables) > 1:  
            controlGroupAvgName = outcomeField + '_' + controlField + 'avg'
            avgTable = 'feat_grpd'+ nameSuffix + '$' + '$'.join(nameParts[1:3]) + '$' + controlGroupAvgName + '$' + '$'.join(nameParts[4:])
            drop = """DROP TABLE IF EXISTS %s""" % (avgTable)
            sql = "create table %s like %s" % (avgTable, newTables[0])
            mm.execute(self.corpdb, self.dbCursor, drop, charset=self.encoding, use_unicode=self.use_unicode)
            mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
            mm.standardizeTable(self.corpdb, self.dbCursor, avgTable, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)
            #create insert fields:
            shortNames = [chr(ord('a')+i) for i in range(len(newTables))]
            tableNames = ', '.join(["%s as %s" % (newTables[i], shortNames[i]) for i in range(len(newTables))])
            values = "(%s)" % (' + '.join(["%s.value" % (name) for name in shortNames])) + ' / ' + str(len(shortNames))
            groupNorms = "(%s)" % (' + '.join(["%s.group_norm" % (name) for name in shortNames])) + ' / ' + str(len(shortNames))
            Ns = ' + '.join(["%s.N" % (name) for name in shortNames])
            stdDev = "(%s)" % (' + '.join(["POW(%s.group_norm - (%s), 2) + POW(%s.std_dev, 2)" % (name, groupNorms, name) for name in shortNames])) + ' / ' + str(len(shortNames))
            stdDev = "SQRT(%s)" % stdDev
            #stdDev = "SQRT(%s)" % (' + '.join(map(lambda name: "%s.N*(POW(%s.group_norm - %s, 2) + POW(%s.std_dev,2))" % (name, name, groupNorms, name), shortNames))) + ' / ' + Ns
            #stdDev = "(%s)" % (' + '.join(map(lambda name: "%s.std_dev_no_zero" % (name), shortNames))) + ' / ' + str(len(shortNames))


            #create joins
            groupIds = ["%s.group_id" % (name) for name in shortNames]
            feats = ["%s.feat" % (name) for name in shortNames]
            groupIdJoins = []
            featJoins = []
            for i in range(len(groupIds) - 1):
                groupIdJoins.append('%s = %s' % (groupIds[i], groupIds[i+1]))
                featJoins.append('%s = %s' % (feats[i], feats[i+1]))
            groupIdJoins = ' AND '.join(groupIdJoins)
            featJoins = ' AND '.join(featJoins)

            #call SQL
            sql = "INSERT INTO %s (group_id, feat, value, group_norm, std_dev, N) SELECT a.group_id, a.feat, %s, %s, %s, %s FROM %s where %s AND %s" % \
                (avgTable, values, groupNorms, stdDev, Ns, tableNames, groupIdJoins, featJoins)
            print("Populating AVG table with command: %s" % sql)
            mm.execute(self.corpdb, self.dbCursor, sql, False, charset=self.encoding, use_unicode=self.use_unicode)


    def createAggregateFeatTableByGroup(self, valueFunc = lambda d: d):
        """combines feature tables, and groups by the given group field"""
        
        featureTable = self.featureTable

        (_, name, oldCorpTable, oldGroupField) = featureTable.split('$')[:4]
        theRest = featureTable.split('$')[4:]

        

        newTable = 'feat$agg_'+name[:12]+'$'+oldCorpTable+'$'+self.correl_field # +'$'+'$'.join(theRest)
        drop = """DROP TABLE IF EXISTS %s""" % (newTable)
        mm.execute(self.corpdb, self.dbCursor, drop, charset=self.encoding, use_unicode=self.use_unicode)

        sql = """CREATE TABLE %s like %s""" % (newTable, featureTable)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
        sql = """ALTER TABLE %s MODIFY group_id VARCHAR(255)""" % (newTable)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
        mm.standardizeTable(self.corpdb, self.dbCursor, newTable, collate=dlac.DEF_COLLATIONS[self.encoding.lower()], engine=dlac.DEF_MYSQL_ENGINE, charset=self.encoding, use_unicode=self.use_unicode)

        mm.disableTableKeys(self.corpdb, self.dbCursor, newTable, charset=self.encoding, use_unicode=self.use_unicode)
        
        dlac.warn("Inserting group_id, feat, and values")
        sql = "INSERT INTO %s SELECT m.%s, f.feat, sum(f.value), 0 FROM %s AS f, %s AS m where m.%s = f.group_id GROUP BY m.%s, f.feat" % (newTable,self.correl_field, featureTable, self.corptable, oldGroupField, self.correl_field)
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)

        dlac.warn("Recalculating group_norms")
        sql = "UPDATE %s a INNER JOIN (SELECT group_id,sum(value) sum FROM %s GROUP BY group_id) b ON a.group_id=b.group_id SET a.group_norm=a.value/b.sum" % (newTable,newTable)
        
        mm.execute(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)
  
        dlac.warn("Done inserting.\nEnabling keys.")
        mm.enableTableKeys(self.corpdb, self.dbCursor, newTable, charset=self.encoding, use_unicode=self.use_unicode)
        dlac.warn("done.")

        self.featureTable = newTable
        return newTable

    def createTfIdfTable(self, ngram_table):
        '''
        Creates new feature table where group_norm = tf-idf (term frequency-inverse document frequency)
        :param ngram_table: table containing words/ngrams, collocs, etc...

        Written by Phil
        '''

        # tf-idf = tf*idf

        # tf (term frequency) is simply how frequently a term occurs in a document (group_norm for a given group_id)

        # each feat's idf = log(N/dt)
        # N = number of documents in total (i.e. count(distinct(group_id))
        # df (document frequency) = number of documents where feat was used in (i.e. count(distinct(group_id)) where feat = 'feat')

        # create new feature table
        feat_name_grabber = re.compile(r'^feat\$([^\$]+)\$') 
        feat_name = feat_name_grabber.match(ngram_table).group(1) # grabs feat_name (i.e. 1gram, 1to3gram)

        short_name = 'tf_idf_{}'.format(feat_name)
        idf_table = self.createFeatureTable(short_name, valueType = 'DOUBLE')

        #getting N
        sql = "SELECT COUNT(DISTINCT group_id) FROM %s" % ngram_table
        N = mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)[0][0]

        feat_counts = self.getFeatureCounts() #tuples of: feat, count (number of groups feature appears with)

        dlac.warn('Inserting idf values into new table')
        counter = 0
        rows = []
        for (feat, dt) in feat_counts:
            idf = log(N/float(dt))

            # get (group_id, group_norm) where feat = feat
            # clean_feat = mm.MySQLdb.escape_string(feat.encode('utf-8')) 

            sql = """SELECT group_id, value, group_norm from %s WHERE feat = \'%s\'"""%(ngram_table, mm.MySQLdb.escape_string(feat.encode('utf-8')).decode('utf-8'))

            group_id_freq = mm.executeGetList(self.corpdb, self.dbCursor, sql, warnQuery=False, charset=self.encoding, use_unicode=self.use_unicode)

            for (group_id, value, tf) in group_id_freq:
                tf_idf = tf * idf

                rows.append([group_id, mm.MySQLdb.escape_string(feat.encode('utf-8')).decode('utf-8'), value, tf_idf])
                if len(rows) >= dlac.MYSQL_BATCH_INSERT_SIZE:
                    sql = "INSERT INTO %s (group_id, feat, value, group_norm) " % idf_table
                    sql += "VALUES (%s)" % ', '.join('%s' for r in rows[0]) 
                    mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)
                    rows = []

                if (counter % 50000 == 0):
                    print('%d tf_idf values inserted!' % (counter))
                counter += 1


        if rows:
            sql = "INSERT INTO %s (group_id, feat, value, group_norm) " % idf_table
            sql += "VALUES (%s)" % ', '.join('%s' for r in rows[0]) 
            mm.executeWriteMany(self.corpdb, self.dbCursor, sql, rows, writeCursor=self.dbConn.cursor(), charset=self.encoding, use_unicode=self.use_unicode)

        dlac.warn('Finished inserting.')
        return idf_table

    def _calc_pmi_iter(self, multigram_counts_iter, onegram_counts_iter, total_count, tokenize_func=lambda x:x.split()):
        '''
        :param multigram_counts_iter: iterator yields dicts, dicts have columns "feat" and "count"
            this is the list of multigrams for which we are calculating pmi values
        :param onegram_counts_iter: iterator yields dicts, dicts have columns "feat" and "count"
            this should include all onegrams that are included in the multigrams from multigram_counts_iter
        :param total_count: the total count of words in the body of text from which onegram and multigrams were generated
        :param tokenize_func: a function that take a string and output's a list of words or tokens
        :return: an iterator of multigrams with their associated PMI information
        '''

        onegram_counts_dict = {}
        print("Reading onegrams into a dict...")
        count = 0
        for onegram_count_row in onegram_counts_iter:
            count += 1
            if count % 100000 == 0:
                print("Processing onegram row {}".format(count))
            onegram = onegram_count_row["feat"]
            onegram_counts_dict[onegram] = onegram_count_row["count"]

        for multigram_count_row in multigram_counts_iter:
            multigram_count_row = dict(multigram_count_row)
            multigram = multigram_count_row['feat']
            onegrams = tokenize_func(multigram)
            num_tokens = len(onegrams)

            onegram_presence = [onegram in onegram_counts_dict for onegram in onegrams]
            if num_tokens <= 1 or not all(onegram_presence):
                (pmi_val, npmi_val, pmi_filter_val, npmi_filter_val) = (None, None, None, None)

            else:
                count_multigram = multigram_count_row['count']
                onegram_counts = [onegram_counts_dict[onegram] for onegram in onegrams]
                pmi_val = self._colloc_pmi(count_multigram, onegram_counts, total_count)
                npmi_val = self._colloc_pmi(multigram_count_row['count'], onegram_counts, total_count, normalize=True)
                pmi_filter_val = pmi_val/(num_tokens - 1)
                npmi_filter_val = npmi_val/(num_tokens - 1)

            multigram_count_row.update({'pmi':pmi_val, 'pmi_filter_val':pmi_filter_val, 'npmi':npmi_val, 'npmi_filter_val':npmi_filter_val})
            yield multigram_count_row

    def _colloc_pmi(self, count_colloc, counts_onegrams, total_count, normalize = False, useAndyDenom = False):
        '''
        Source: https://svn.spraakdata.gu.se/repos/gerlof/pub/www/Docs/npmi-pfd.pdf
        :param prob_colloc: float from 0 to 1
        :param prob_ngrams: list of floats from 0 to 1
        :return: float from ??? to ???
        '''
        prob_colloc = float(count_colloc)/float(total_count)
        probs_onegrams = [float(onegram_count)/float(total_count) for onegram_count in counts_onegrams]
        if useAndyDenom:
            raise NotImplementedError
        else:
            denom = float(reduce(mul, probs_onegrams, 1))

        ratio = prob_colloc/denom
        pmi_val = math.log(ratio, math.e)
        if normalize:
            normalization_factor = -1 * math.log(prob_colloc)
            npmi_val = pmi_val/normalization_factor
            return npmi_val
        else:
            return pmi_val

    def _colloc_lnpmi_vals(self, colloc_list, counts, total_word_count):
        '''
        :param colloc_list - list of collocs that we want values calculated for
        :param counts - dict keyed by colloc, where multiple tokens are separated by spaces
        :param total_word_count - int
        :returns dict iter
        '''
        for colloc in colloc_list:
            ##TODO - find way to deal with missing onegrams or two grams
            onegrams = colloc.split()
            num_tokens = len(onegrams)
            prob_colloc = float(counts[colloc])/float(total_word_count)
            normalization_factor = -1 * math.log(prob_colloc)
            onegram_probs = [float(counts[onegram])/float(total_word_count) for onegram in onegrams]
            pmi_denom = reduce(mul, onegram_probs, 1)
            if num_tokens >= 3:
                try:
                    count_prod1 = counts[onegrams[0]] * counts[" ".join(onegrams[1:])]
                    count_prod2 = counts[" ".join(onegrams[:-1])] * counts[onegrams[-1]]
                except Exception as e:
                    raise e
                andy_denom = float(max(count_prod1, count_prod2))/math.pow(total_word_count, 2)
            pmi_val = math.log(prob_colloc/pmi_denom, math.e)
            lpmi_val = math.log(prob_colloc/andy_denom, math.e) if num_tokens >= 3 else pmi_val
            npmi_val = pmi_val/normalization_factor
            lnpmi_val = lpmi_val/normalization_factor
            yield {'feat':colloc, 'num_tokens':num_tokens, 'pmi':pmi_val, 'npmi':npmi_val, 'lpmi':lpmi_val, 'lnpmi':lnpmi_val}


    def creatCollocScores(self, ufeat_table):
        """"""
        if not isinstance(ufeat_table, str):
            ufeat_table = self.corptable
        db_eng = mif.get_db_engine(self.corpdb)

        ufeat_multigram_table = "ufeat$" + ufeat_table

        drop_sql = "DROP TABLE IF EXISTS {ufeat}".format(ufeat=ufeat_multigram_table)
        dlac.warn(drop_sql)
        db_eng.execute(drop_sql)
        create_sql = """CREATE TABLE {ufeat}
            (id BIGINT PRIMARY KEY AUTO_INCREMENT, feat varchar(102), count bigint, KEY feat (feat) ) 
            DEFAULT CHARSET=utf8mb4""".format(ufeat=ufeat_multigram_table)
        dlac.warn(create_sql)
        db_eng.execute(create_sql)
        
        insert_sql = """INSERT INTO {ufeat} (feat, count) SELECT feat, sum(value) count 
            FROM {ftbl} GROUP BY feat""".format(ufeat=ufeat_multigram_table, ftbl=self.featureTable)
        dlac.warn(insert_sql)
        db_eng.execute(insert_sql)

        print("Extending table if necessary...")
        new_cols = OrderedDict()
        new_cols['pmi'] = 'DOUBLE';
        new_cols['pmi_filter_val'] = 'DOUBLE';
        new_cols['npmi'] = 'DOUBLE';
        new_cols['npmi_filter_val'] = 'DOUBLE';
        new_cols['npmi'] = 'DOUBLE';
        new_cols['npmi_filter_val'] = 'DOUBLE';
        group_column = self.featureTable.split('$')[3]
        pocc_column = "pocc_{}_gft0".format(group_column)
        new_cols[pocc_column] = 'DOUBLE';
        mif.extend_table(db_eng, ufeat_multigram_table, new_cols)

        print("Querying input data...")
        total_count = db_eng.execute("SELECT sum(value) FROM {}".format(self.wordTable)).first()[0]

        ###AHHHH this MUST be grouped!!!!
        onegram_counts_iter =  db_eng.execute("SELECT id, feat, SUM(value) as count FROM {} GROUP BY feat".format(self.wordTable))
        multigram_counts_iter =  db_eng.execute("SELECT id, feat, count FROM {} WHERE pmi IS NULL AND feat LIKE '% %' AND count > 1".format(ufeat_multigram_table))

        pmi_iter = self._calc_pmi_iter(multigram_counts_iter, onegram_counts_iter, total_count)

        print("Processing npmi data...")
        mif.mysql_update(db_eng, ufeat_multigram_table, pmi_iter)

        print("Done npmi.")

        ### annotate pocc
        num_groups_tot = db_eng.execute("SELECT count(distinct group_id) FROM {}".format(self.featureTable)).first()[0]

        print("Loading group counts by feat...")
        group_count_sql = "SELECT feat, count(*) group_count FROM {} GROUP BY feat".format(self.featureTable)
        df_group_counts = pd.read_sql(group_count_sql, db_eng, index_col="feat")

        print("Loading ufeat data...")
        feat_iter =  db_eng.execute("SELECT id, feat FROM {} WHERE count > 1".format(ufeat_multigram_table))

        def pocc_gen():
            count = 0
            for (id, feat) in feat_iter:
                num_groups = df_group_counts.ix[feat]['group_count']
                yield {'id':id, pocc_column:float(num_groups)/num_groups_tot}
        pocc_dict_iter = pocc_gen()

        print("Updating ufeat data with pocc values...")
        mif.mysql_update(db_eng, ufeat_multigram_table, pocc_dict_iter, log_every=10000)
        return ufeat_multigram_table



"""
Semantic Extractor

Bridges DLATK's Feature Extractor with reading xml semantic annotations
assumes that corptable is a directory with files rather than a database table. 
"""
import os
import gzip
import sys
import xml.etree.cElementTree as ET

from pprint import pprint

from .featureExtractor import FeatureExtractor
from .outcomeGetter import OutcomeGetter
from . import dlaWorker

#############################################################
## Static Variables and Methods
#
COLLOC_LENGTH = 96

def _warn(string):
    print(string, file=sys.stderr)

class SemanticsExtractor(FeatureExtractor):

    def __init__(self, corpdb, corptable, correl_field, mysql_host, message_field, messageid_field, corpdir):
        #corptable = "_".join(corpdir.split('/')[-2:])
        super(FeatureExtractor, self).__init__(corpdb, corptable, correl_field, mysql_host, message_field, messageid_field)
        self.corpdir = corpdir


    def addNERTable(self, tableName = None, min_freq=1, valueFunc = lambda d: d, normalizeByCollocsInGroup = False):
        """extracts named entiry features from semantic xml files"""
        #assumes each file is one group
        #normalizeByCollocsInGroup: if true, will use the number of collocs as the value to normalize by rather than terms 

        ##CREATE TABLES##
        tableNames = {'ner': self.createFeatureTable('ner', "VARCHAR(%d)"%int(COLLOC_LENGTH/3), 'INTEGER', tableName, valueFunc),
                      'ner_colloc': self.createFeatureTable('ner_colloc', "VARCHAR(%d)"%int(COLLOC_LENGTH*1.333), 'INTEGER', tableName, valueFunc),
                      'colloc':  self.createFeatureTable('colloc', "VARCHAR(%d)"%COLLOC_LENGTH, 'INTEGER', tableName, valueFunc)}

        ##FIND FILES##
        corpdir = self.corpdir
        subdirs = os.listdir(corpdir)
        groupsInserted = 0
        if len(subdirs) > 50: [self._disableTableKeys(name) for name in list(tableNames.values())]#for faster, when enough space for repair by sorting
        for subdir in subdirs:
            #print "on %s" % subdir #debug

            files = os.listdir(corpdir+'/'+subdir)
            for filename in files:
                f = gzip.open(corpdir+'/'+subdir+'/'+filename, 'rb')
                xmlStr = f.read()
                f.close()
                (_, _, xmlStr) = xmlStr.partition("\n")
                xmlRoot = None
                try: 
                    xmlRoot = ET.fromstring(xmlStr)
                except ET.ParseError:
                    #skip if xml is malformed
                    continue

                #setup group specific variables
                freqs = {'ner': dict(),
                         'colloc': dict(),
                         'ner_colloc': dict()}
                totalTermsInGroup = 0
                totalCollocsInGroup = 0
                for document in xmlRoot.findall('document'):
                    for parsed in document.findall('parsed'):
                        for section in parsed.findall('section'):
                            for sent in section.findall('sentence'):
                                #at this point we have terms and nonterms, nonterms group terms together
                                for nonterm in sent.findall('nonterm'):
                                    #add to totalTokens
                                    totalTermsInGroup += len(nonterm.findall('term'))
                                    totalCollocsInGroup += 1
                                    #get ne class + colloc
                                    (ne, colloc) = ('','')
                                    for attr in nonterm.findall('attr'):
                                        if attr.attrib['name'] == 'ne':
                                            ne = attr.attrib['value']



                                        elif attr.attrib['name'] == 'nealias':
                                            colloc = attr.attrib['value']
                                    try:
                                        freqs['ner'][ne] += 1
                                    except KeyError:
                                        freqs['ner'][ne] = 1
                                    try:
                                        freqs['colloc'][colloc] += 1
                                    except KeyError:
                                        freqs['colloc'][colloc] = 1
                                    try:
                                        freqs['ner_colloc']["%s:%s"%(ne,colloc)] += 1
                                    except KeyError:
                                        freqs['ner_colloc']["%s:%s"%(ne,colloc)] = 1

                                totalTermsInGroup += len(nonterm.findall('term'))
                                    
                #pprint(freqs['ner_colloc']) #debug
                #print totalTermsInGroup #debug

                ##INSERT INTO DB##
                group_id = filename.split('.')[0] #first part of filename is group id
                #print group_id #debug
                total = float(totalTermsInGroup)
                if normalizeByCollocsInGroup:
                    total = float(totalCollocsInGroup)
                for dataType, featFreqs in freqs.items():
                    wsql = """INSERT INTO """+tableNames[dataType]+""" (group_id, feat, value, group_norm) values ('"""+str(group_id)+"""', %s, %s, %s)"""
                    rows = [(k.encode('utf-8'), v, valueFunc((v / total))) for k, v in featFreqs.items() if v >= min_freq] #adds group_norm and applies freq filter
                    self._executeWriteMany(wsql, rows)

                groupsInserted+=1
                if groupsInserted % 1000 == 0:
                    _warn("\n\t%d groups inserted" % groupsInserted)

                #raise NotImplementedError

        if len(subdirs) > 50: 
            _warn("Adding Keys (if goes to keycache, then consider running myisamchk -n).")
            [self._enableTableKeys(name) for name in list(tableNames.values())]#for faster, when enough space for repair by sorting

        return tableNames['ner'] #feature creation methods are expected to return 1 table name so I just chose NER
